import logging
import os
import unittest
from datetime import datetime, timedelta
from email.utils import formataddr
from time import sleep

import requests
from django.test import SimpleTestCase, override_settings, tag

from anymail.exceptions import AnymailAPIError
from anymail.message import AnymailMessage

from .utils import AnymailTestMixin, sample_image_path

ANYMAIL_TEST_MAILGUN_API_KEY = os.getenv("ANYMAIL_TEST_MAILGUN_API_KEY")
ANYMAIL_TEST_MAILGUN_DOMAIN = os.getenv("ANYMAIL_TEST_MAILGUN_DOMAIN")


@tag("mailgun", "live")
@unittest.skipUnless(
    ANYMAIL_TEST_MAILGUN_API_KEY and ANYMAIL_TEST_MAILGUN_DOMAIN,
    "Set ANYMAIL_TEST_MAILGUN_API_KEY and ANYMAIL_TEST_MAILGUN_DOMAIN environment"
    " variables to run Mailgun integration tests",
)
@override_settings(
    ANYMAIL={
        "MAILGUN_API_KEY": ANYMAIL_TEST_MAILGUN_API_KEY,
        "MAILGUN_SENDER_DOMAIN": ANYMAIL_TEST_MAILGUN_DOMAIN,
        "MAILGUN_SEND_DEFAULTS": {"esp_extra": {"o:testmode": "yes"}},
    },
    EMAIL_BACKEND="anymail.backends.mailgun.EmailBackend",
)
class MailgunBackendIntegrationTests(AnymailTestMixin, SimpleTestCase):
    """Mailgun API integration tests

    These tests run against the **live** Mailgun API, using the
    environment variable `ANYMAIL_TEST_MAILGUN_API_KEY` as the API key
    and `ANYMAIL_TEST_MAILGUN_DOMAIN` as the sender domain.
    If those variables are not set, these tests won't run.

    """

    def setUp(self):
        super().setUp()
        self.from_email = "from@%s" % ANYMAIL_TEST_MAILGUN_DOMAIN
        self.message = AnymailMessage(
            "Anymail Mailgun integration test",
            "Text content",
            self.from_email,
            ["test+to1@anymail.dev"],
        )
        self.message.attach_alternative("<p>HTML content</p>", "text/html")

    def fetch_mailgun_events(
        self, message_id, event=None, initial_delay=2, retry_delay=2, max_retries=5
    ):
        """Return list of Mailgun events related to message_id"""
        url = "https://api.mailgun.net/v3/%s/events" % ANYMAIL_TEST_MAILGUN_DOMAIN
        auth = ("api", ANYMAIL_TEST_MAILGUN_API_KEY)

        # Despite the docs, Mailgun's events API actually expects the message-id
        # without the <...> brackets (so, not exactly "as returned by the messages API")
        # https://documentation.mailgun.com/api-events.html#filter-field
        params = {"message-id": message_id[1:-1]}  # strip <...>
        if event is not None:
            params["event"] = event

        # It can take a few seconds for the events to show up
        # in Mailgun's logs, so retry a few times if necessary:
        sleep(initial_delay)
        response = None
        for retry in range(max_retries):
            if retry > 0:
                sleep(retry_delay)
            response = requests.get(url, auth=auth, params=params)
            if 200 == response.status_code:
                items = response.json()["items"]
                if len(items) > 0:
                    return items
                # else no matching events found yet, so try again after delay
            elif 500 <= response.status_code < 600:
                # server error (hopefully transient); try again after delay
                pass
            elif 403 == response.status_code:
                # "forbidden": this may be related to API throttling;
                # try again after delay
                pass
            else:
                response.raise_for_status()
        # Max retries exceeded:
        if response is not None and 200 != response.status_code:
            logging.warning(
                "Ignoring Mailgun events API error %d:\n%s"
                % (response.status_code, response.text)
            )
        return None

    def test_simple_send(self):
        # Example of getting the Mailgun send status and message id from the message
        sent_count = self.message.send()
        self.assertEqual(sent_count, 1)

        anymail_status = self.message.anymail_status
        sent_status = anymail_status.recipients["test+to1@anymail.dev"].status
        message_id = anymail_status.recipients["test+to1@anymail.dev"].message_id

        self.assertEqual(sent_status, "queued")  # Mailgun always queues
        # don't know what it'll be, but it should exist:
        self.assertGreater(len(message_id), 0)

        # set of all recipient statuses:
        self.assertEqual(anymail_status.status, {sent_status})
        self.assertEqual(anymail_status.message_id, message_id)

    def test_all_options(self):
        send_at = datetime.now().replace(microsecond=0) + timedelta(minutes=2)
        send_at_timestamp = send_at.timestamp()
        from_email = formataddr(("Test From, with comma", self.from_email))
        message = AnymailMessage(
            subject="Anymail Mailgun all-options integration test",
            body="This is the text body",
            from_email=from_email,
            to=["test+to1@anymail.dev", "Recipient 2 <test+to2@anymail.dev>"],
            cc=["test+cc1@anymail.dev", "Copy 2 <test+cc2@anymail.dev>"],
            bcc=["test+bcc1@anymail.dev", "Blind Copy 2 <test+bcc2@anymail.dev>"],
            reply_to=["reply1@example.com", "Reply 2 <reply2@example.com>"],
            headers={"X-Anymail-Test": "value"},
            metadata={"meta1": "simple string", "meta2": 2},
            send_at=send_at,
            tags=["tag 1", "tag 2"],
            track_clicks=False,
            track_opens=True,
        )
        message.attach("attachment1.txt", "Here is some\ntext for you", "text/plain")
        message.attach("vedhæftet fil.csv", "ID,Name\n1,3", "text/csv")
        cid = message.attach_inline_image_file(
            sample_image_path(), domain=ANYMAIL_TEST_MAILGUN_DOMAIN
        )
        message.attach_alternative(
            "<div>This is the <i>html</i> body <img src='cid:%s'></div>" % cid,
            "text/html",
        )

        message.send()
        # Mailgun always queues:
        self.assertEqual(message.anymail_status.status, {"queued"})
        message_id = message.anymail_status.message_id

        events = self.fetch_mailgun_events(message_id, event="accepted")
        if events is None:
            self.skipTest(
                "No Mailgun 'accepted' event after 30sec -- can't complete this test"
            )
            return

        event = events.pop()
        # don't care about order:
        self.assertCountEqual(event["tags"], ["tag 1", "tag 2"])
        # all metadata values become strings:
        self.assertEqual(
            event["user-variables"], {"meta1": "simple string", "meta2": "2"}
        )

        self.assertEqual(event["message"]["scheduled-for"], send_at_timestamp)
        self.assertIn(
            event["recipient"],
            [
                "test+to1@anymail.dev",
                "test+to2@anymail.dev",
                "test+cc1@anymail.dev",
                "test+cc2@anymail.dev",
                "test+bcc1@anymail.dev",
                "test+bcc2@anymail.dev",
            ],
        )

        headers = event["message"]["headers"]
        self.assertEqual(headers["from"], from_email)
        self.assertEqual(
            headers["to"], "test+to1@anymail.dev, Recipient 2 <test+to2@anymail.dev>"
        )
        self.assertEqual(
            headers["subject"], "Anymail Mailgun all-options integration test"
        )

        attachments = event["message"]["attachments"]
        if len(attachments) == 3:
            # The inline attachment shouldn't be in the event message.attachments array,
            # but sometimes is included for the accepted event (see #172)
            inline_attachment = attachments.pop(0)
            self.assertEqual(inline_attachment["filename"], cid)
            self.assertEqual(inline_attachment["content-type"], "image/png")
        self.assertEqual(len(attachments), 2)
        self.assertEqual(attachments[0]["filename"], "attachment1.txt")
        self.assertEqual(attachments[0]["content-type"], "text/plain")
        self.assertEqual(attachments[1]["filename"], "vedhæftet fil.csv")
        self.assertEqual(attachments[1]["content-type"], "text/csv")

        # No other fields are verifiable from the event data.
        # (We could try fetching the message from event["storage"]["url"]
        # to verify content and other headers.)

    def test_per_recipient_options(self):
        message = AnymailMessage(
            from_email=formataddr(("Test From", self.from_email)),
            to=["test+to1@anymail.dev", '"Recipient 2" <test+to2@anymail.dev>'],
            subject="Anymail Mailgun per-recipient options test",
            body="This is the text body",
            merge_metadata={
                "test+to1@anymail.dev": {"meta1": "one", "meta2": "two"},
                "test+to2@anymail.dev": {"meta1": "recipient 2"},
            },
            headers={
                "List-Unsubscribe-Post": "List-Unsubscribe=One-Click",
                "List-Unsubscribe": "<mailto:unsubscribe@example.com>",
                "X-Custom-Header": "default",
            },
            merge_headers={
                "test+to1@anymail.dev": {
                    "List-Unsubscribe": "<https://example.com/a/>",
                    "X-Custom-Header": "custom",
                },
                "test+to2@anymail.dev": {
                    "List-Unsubscribe": "<https://example.com/b/>",
                },
            },
        )
        message.send()
        recipient_status = message.anymail_status.recipients
        self.assertEqual(recipient_status["test+to1@anymail.dev"].status, "queued")
        self.assertEqual(recipient_status["test+to2@anymail.dev"].status, "queued")

    def test_stored_template(self):
        message = AnymailMessage(
            # name of a real template named in Anymail's Mailgun test account:
            template_id="test-template",
            # Mailgun templates don't define subject:
            subject="Your order %recipient.order%",
            # Mailgun templates don't define sender:
            from_email=formataddr(("Test From>", self.from_email)),
            to=["test+to1@anymail.dev"],
            # metadata and merge_data must not have any conflicting keys
            # when using template_id:
            metadata={"meta1": "simple string", "meta2": 2},
            merge_data={
                "test+to1@anymail.dev": {
                    "name": "Test Recipient",
                }
            },
            merge_global_data={
                "order": "12345",
            },
        )
        message.send()
        recipient_status = message.anymail_status.recipients
        self.assertEqual(recipient_status["test+to1@anymail.dev"].status, "queued")

    # As of Anymail 0.10, this test is no longer possible, because
    # Anymail now raises AnymailInvalidAddress without even calling Mailgun
    # def test_invalid_from(self):
    #     self.message.from_email = 'webmaster'
    #     with self.assertRaises(AnymailAPIError) as cm:
    #         self.message.send()
    #     err = cm.exception
    #     self.assertEqual(err.status_code, 400)
    #     self.assertIn("'from' parameter is not a valid address", str(err))

    @override_settings(
        ANYMAIL={
            "MAILGUN_API_KEY": "Hey, that's not an API key",
            "MAILGUN_SENDER_DOMAIN": ANYMAIL_TEST_MAILGUN_DOMAIN,
            "MAILGUN_SEND_DEFAULTS": {"esp_extra": {"o:testmode": "yes"}},
        }
    )
    def test_invalid_api_key(self):
        with self.assertRaises(AnymailAPIError) as cm:
            self.message.send()
        err = cm.exception
        self.assertEqual(err.status_code, 401)
        # Mailgun doesn't offer any additional explanation in its response body
        # self.assertIn("Forbidden", str(err))



import os
import unittest

from django.test import SimpleTestCase, override_settings, tag

from anymail.exceptions import AnymailAPIError
from anymail.message import AnymailMessage

from .utils import AnymailTestMixin

ANYMAIL_TEST_SCALEWAY_SECRET_KEY = os.getenv("ANYMAIL_TEST_SCALEWAY_SECRET_KEY")
ANYMAIL_TEST_SCALEWAY_PROJECT_ID = os.getenv("ANYMAIL_TEST_SCALEWAY_PROJECT_ID")
ANYMAIL_TEST_SCALEWAY_DOMAIN = os.getenv("ANYMAIL_TEST_SCALEWAY_DOMAIN")


@tag("scaleway", "live")
@unittest.skipUnless(
    ANYMAIL_TEST_SCALEWAY_SECRET_KEY
    and ANYMAIL_TEST_SCALEWAY_PROJECT_ID
    and ANYMAIL_TEST_SCALEWAY_DOMAIN,
    "Set ANYMAIL_TEST_SCALEWAY_SECRET_KEY, ANYMAIL_TEST_SCALEWAY_PROJECT_ID, and "
    "ANYMAIL_TEST_SCALEWAY_DOMAIN environment variables to run Scaleway integration tests",
)
@override_settings(
    ANYMAIL={
        "SCALEWAY_SECRET_KEY": ANYMAIL_TEST_SCALEWAY_SECRET_KEY,
        "SCALEWAY_PROJECT_ID": ANYMAIL_TEST_SCALEWAY_PROJECT_ID,
    },
    EMAIL_BACKEND="anymail.backends.scaleway.EmailBackend",
)
class ScalewayIntegrationTests(AnymailTestMixin, SimpleTestCase):
    """Scaleway API integration tests

    These tests run against the **live** Scaleway API, using the
    environment variables to authorize.
    If those variables are not set, these tests will be skipped.
    """

    esp_name = "Scaleway"

    def setUp(self):
        super().setUp()
        self.from_email = f"from@{ANYMAIL_TEST_SCALEWAY_DOMAIN}"
        self.message = AnymailMessage(
            subject="Anymail Scaleway integration test",
            body="This is a test message from Anymail.",
            from_email=self.from_email,
            to=["test+to1@anymail.dev"],
        )
        self.message.attach_alternative("<p>HTML content</p>", "text/html")

    def test_simple_send(self):
        """Test basic sending."""
        self.message.send()
        self.assertEqual(self.message.anymail_status.status, {"queued"})
        self.assertIsNotNone(self.message.anymail_status.message_id)

    def test_all_options(self):
        """Test sending with all available options."""
        message = AnymailMessage(
            subject="Anymail Scaleway all-options test",
            body="This is a test message from Anymail.",
            from_email=f"from@{ANYMAIL_TEST_SCALEWAY_DOMAIN}",
            to=["test+to1@anymail.dev", '"Recipient 2, OK?" <test+to2@anymail.dev>'],
            cc=["test+cc1@anymail.dev", "Copy 2 <test+cc2@anymail.dev>"],
            bcc=["test+bcc1@anymail.dev", "Blind Copy 2 <test+bcc2@anymail.dev>"],
            reply_to=["reply1@example.com", "Reply 2 <reply2@example.com>"],
            headers={"X-Anymail-Test": "all-options", "X-Anymail-Count": 3},
            metadata={"meta1": "simple string", "meta2": 2},
            tags=["tag 1", "tag 2"],
        )
        message.attach_alternative("<p>HTML content</p>", "text/html")
        message.attach("attachment1.txt", "Here is some\ntext for you", "text/plain")
        message.attach("attachment2.csv", "ID,Name\n1,Amy Lina", "text/csv")

        message.send()
        self.assertEqual(len(message.anymail_status.recipients), 6)
        self.assertEqual(message.anymail_status.status, {"queued"})
        self.assertIsNotNone(message.anymail_status.message_id)

    def test_invalid_from(self):
        self.message.from_email = "webmaster@localhost"
        with self.assertRaises(AnymailAPIError) as cm:
            self.message.send()
        err = cm.exception
        self.assertEqual(err.status_code, 400)
        self.assertIn("Invalid email from address", str(err))

    @override_settings(
        ANYMAIL={
            "SCALEWAY_SECRET_KEY": "Hey, that's not a secret key",
            "SCALEWAY_PROJECT_ID": "not-a-project-id",
        }
    )
    def test_invalid_secret_key(self):
        with self.assertRaises(AnymailAPIError) as cm:
            self.message.send()
        err = cm.exception
        self.assertEqual(err.status_code, 401)
        self.assertIn("authentication is denied", str(err))



from django.core import checks
from django.test import SimpleTestCase
from django.test.utils import override_settings

from anymail.checks import check_deprecated_settings, check_insecure_settings

from .utils import AnymailTestMixin


class DeprecatedSettingsTests(AnymailTestMixin, SimpleTestCase):
    @override_settings(ANYMAIL={"WEBHOOK_AUTHORIZATION": "abcde:12345"})
    def test_webhook_authorization(self):
        errors = check_deprecated_settings(None)
        self.assertEqual(
            errors,
            [
                checks.Error(
                    "The ANYMAIL setting 'WEBHOOK_AUTHORIZATION' has been renamed"
                    " 'WEBHOOK_SECRET' to improve security.",
                    hint="You must update your settings.py.",
                    id="anymail.E001",
                )
            ],
        )

    @override_settings(ANYMAIL_WEBHOOK_AUTHORIZATION="abcde:12345", ANYMAIL={})
    def test_anymail_webhook_authorization(self):
        errors = check_deprecated_settings(None)
        self.assertEqual(
            errors,
            [
                checks.Error(
                    "The ANYMAIL_WEBHOOK_AUTHORIZATION setting has been renamed"
                    " ANYMAIL_WEBHOOK_SECRET to improve security.",
                    hint="You must update your settings.py.",
                    id="anymail.E001",
                )
            ],
        )

    @override_settings(EMAIL_BACKEND="anymail.backends.sendgrid.EmailBackend")
    def test_sendgrid_unsupported(self):
        errors = check_deprecated_settings(None)
        self.assertEqual(
            errors,
            [
                checks.Warning(
                    "django-anymail has dropped official support for SendGrid.",
                    hint="See https://github.com/anymail/django-anymail/issues/432.",
                    id="anymail.W003",
                )
            ],
        )


class InsecureSettingsTests(AnymailTestMixin, SimpleTestCase):
    @override_settings(ANYMAIL={"DEBUG_API_REQUESTS": True})
    def test_debug_api_requests_deployed(self):
        errors = check_insecure_settings(None)
        self.assertEqual(len(errors), 1)
        self.assertEqual(errors[0].id, "anymail.W002")

    @override_settings(ANYMAIL={"DEBUG_API_REQUESTS": True}, DEBUG=True)
    def test_debug_api_requests_debug(self):
        errors = check_insecure_settings(None)
        self.assertEqual(len(errors), 0)  # no warning in DEBUG (non-production) config



from textwrap import dedent
from unittest.mock import ANY

from django.test import override_settings, tag

from anymail.inbound import AnymailInboundMessage
from anymail.signals import AnymailInboundEvent
from anymail.webhooks.mandrill import MandrillCombinedWebhookView

from .test_mandrill_webhooks import TEST_WEBHOOK_KEY, mandrill_args
from .webhook_cases import WebhookTestCase


@tag("mandrill")
@override_settings(ANYMAIL_MANDRILL_WEBHOOK_KEY=TEST_WEBHOOK_KEY)
class MandrillInboundTestCase(WebhookTestCase):
    def test_inbound_basics(self):
        raw_event = {
            "event": "inbound",
            "ts": 1507856722,
            "msg": {
                "raw_msg": dedent(
                    """\
                    From: A tester <test@example.org>
                    Date: Thu, 12 Oct 2017 18:03:30 -0700
                    Message-ID: <CAEPk3RKEx@mail.example.org>
                    Subject: Test subject
                    To: "Test, Inbound" <test@inbound.example.com>, other@example.com
                    MIME-Version: 1.0
                    Content-Type: multipart/alternative; boundary="94eb2c05e174adb140055b6339c5"

                    --94eb2c05e174adb140055b6339c5
                    Content-Type: text/plain; charset="UTF-8"
                    Content-Transfer-Encoding: quoted-printable

                    It's a body=E2=80=A6

                    --94eb2c05e174adb140055b6339c5
                    Content-Type: text/html; charset="UTF-8"
                    Content-Transfer-Encoding: quoted-printable

                    <div dir=3D"ltr">It's a body=E2=80=A6</div>

                    --94eb2c05e174adb140055b6339c5--
                    """  # NOQA: E501
                ),
                "email": "delivered-to@example.com",
                # Mandrill populates "sender" only for outbound message events
                "sender": None,
                "spam_report": {
                    "score": 1.7,
                },
                # Anymail ignores Mandrill's other inbound event fields
                # (which are all redundant with raw_msg)
            },
        }

        response = self.client.post(
            **mandrill_args(events=[raw_event], path="/anymail/mandrill/")
        )
        self.assertEqual(response.status_code, 200)
        kwargs = self.assert_handler_called_once_with(
            self.inbound_handler,
            sender=MandrillCombinedWebhookView,
            event=ANY,
            esp_name="Mandrill",
        )
        # Inbound should not dispatch tracking signal:
        self.assertEqual(self.tracking_handler.call_count, 0)

        event = kwargs["event"]
        self.assertIsInstance(event, AnymailInboundEvent)
        self.assertEqual(event.event_type, "inbound")
        self.assertEqual(event.timestamp.isoformat(), "2017-10-13T01:05:22+00:00")
        self.assertIsNone(event.event_id)  # Mandrill doesn't provide inbound event id
        self.assertIsInstance(event.message, AnymailInboundMessage)
        self.assertEqual(event.esp_event, raw_event)

        message = event.message
        self.assertEqual(message.from_email.display_name, "A tester")
        self.assertEqual(message.from_email.addr_spec, "test@example.org")
        self.assertEqual(len(message.to), 2)
        self.assertEqual(message.to[0].display_name, "Test, Inbound")
        self.assertEqual(message.to[0].addr_spec, "test@inbound.example.com")
        self.assertEqual(message.to[1].addr_spec, "other@example.com")
        self.assertEqual(message.subject, "Test subject")
        self.assertEqual(message.date.isoformat(" "), "2017-10-12 18:03:30-07:00")
        self.assertEqual(message.text, "It's a body\N{HORIZONTAL ELLIPSIS}\n")
        self.assertEqual(
            message.html,
            """<div dir="ltr">It's a body\N{HORIZONTAL ELLIPSIS}</div>\n""",
        )

        self.assertIsNone(message.envelope_sender)  # Mandrill doesn't provide sender
        self.assertEqual(message.envelope_recipient, "delivered-to@example.com")
        # Mandrill doesn't provide stripped plaintext body:
        self.assertIsNone(message.stripped_text)
        # Mandrill doesn't provide stripped html:
        self.assertIsNone(message.stripped_html)
        # Mandrill doesn't provide spam boolean:
        self.assertIsNone(message.spam_detected)
        self.assertEqual(message.spam_score, 1.7)

        # Anymail will also parse attachments (if any) from the raw mime.
        # We don't bother testing that here; see test_inbound for examples.



from __future__ import annotations

import re
import typing
import uuid
from datetime import datetime, timezone
from email.charset import QP, Charset
from email.headerregistry import Address

from django.core.mail import EmailMessage
from requests import Response
from requests.structures import CaseInsensitiveDict

from anymail.backends.base_requests import AnymailRequestsBackend, RequestsPayload
from anymail.message import AnymailRecipientStatus
from anymail.utils import Attachment, EmailAddress, get_anymail_setting, update_deep

# Used to force RFC-2047 encoded word
# in address formatting workaround
QP_CHARSET = Charset("utf-8")
QP_CHARSET.header_encoding = QP


class EmailBackend(AnymailRequestsBackend):
    """Unisender Go v1 Web API Email Backend"""

    esp_name = "Unisender Go"

    def __init__(self, **kwargs: typing.Any):
        """Init options from Django settings"""
        esp_name = self.esp_name

        self.api_key = get_anymail_setting(
            "api_key", esp_name=esp_name, kwargs=kwargs, allow_bare=True
        )

        self.generate_message_id = get_anymail_setting(
            "generate_message_id", esp_name=esp_name, kwargs=kwargs, default=True
        )

        # No default for api_url setting -- it depends on account's data center. E.g.:
        # - https://go1.unisender.ru/ru/transactional/api/v1
        # - https://go2.unisender.ru/ru/transactional/api/v1
        api_url = get_anymail_setting("api_url", esp_name=esp_name, kwargs=kwargs)
        if not api_url.endswith("/"):
            api_url += "/"

        # Undocumented setting to control workarounds for Unisender Go display-name issues
        # (see below). If/when Unisender Go fixes the problems, you can disable Anymail's
        # workarounds by adding `"UNISENDER_GO_WORKAROUND_DISPLAY_NAME_BUGS": False`
        # to your `ANYMAIL` settings.
        self.workaround_display_name_bugs = get_anymail_setting(
            "workaround_display_name_bugs",
            esp_name=esp_name,
            kwargs=kwargs,
            default=True,
        )

        super().__init__(api_url, **kwargs)

    def build_message_payload(
        self, message: EmailMessage, defaults: dict
    ) -> UnisenderGoPayload:
        return UnisenderGoPayload(message=message, defaults=defaults, backend=self)

    # Map Unisender Go "failed_email" code -> AnymailRecipientStatus.status
    _unisender_failure_status = {
        # "duplicate": ignored (see parse_recipient_status)
        "invalid": "invalid",
        "permanent_unavailable": "rejected",
        "temporary_unavailable": "failed",
        "unsubscribed": "rejected",
    }

    def parse_recipient_status(
        self, response: Response, payload: UnisenderGoPayload, message: EmailMessage
    ) -> dict:
        """
        Response example:
        {
          "status": "success",
          "job_id": "1ZymBc-00041N-9X",
          "emails": [
            "user@example.com",
            "email@example.com",
          ],
          "failed_emails": {
            "email1@gmail.com": "temporary_unavailable",
            "bad@address": "invalid",
            "email@example.com": "duplicate",
            "root@example.org": "permanent_unavailable",
            "olduser@example.net": "unsubscribed"
          }
        }
        """
        parsed_response = self.deserialize_json_response(response, payload, message)
        # job_id serves as message_id when not self.generate_message_id
        job_id = parsed_response.get("job_id")
        succeed_emails = {
            recipient: AnymailRecipientStatus(
                message_id=payload.message_ids.get(recipient, job_id), status="queued"
            )
            for recipient in parsed_response["emails"]
        }
        failed_emails = {
            recipient: AnymailRecipientStatus(
                # Message wasn't sent to this recipient, so Unisender Go hasn't stored
                # any metadata (including message_id)
                message_id=None,
                status=self._unisender_failure_status.get(status, "failed"),
            )
            for recipient, status in parsed_response.get("failed_emails", {}).items()
            if status != "duplicate"  # duplicates are in both succeed and failed lists
        }
        return {**succeed_emails, **failed_emails}


class UnisenderGoPayload(RequestsPayload):
    # Payload: see https://godocs.unisender.ru/web-api-ref#email-send

    data: dict

    def __init__(
        self,
        message: EmailMessage,
        defaults: dict,
        backend: EmailBackend,
        *args: typing.Any,
        **kwargs: typing.Any,
    ):
        self.generate_message_id = backend.generate_message_id
        self.message_ids = CaseInsensitiveDict()  # recipient -> generated message_id

        http_headers = kwargs.pop("headers", {})
        http_headers["Content-Type"] = "application/json"
        http_headers["Accept"] = "application/json"
        http_headers["X-API-KEY"] = backend.api_key
        super().__init__(
            message, defaults, backend, headers=http_headers, *args, **kwargs
        )

    def get_api_endpoint(self) -> str:
        return "email/send.json"

    def init_payload(self) -> None:
        self.data = {  # becomes json
            "headers": CaseInsensitiveDict(),
            "recipients": [],
        }

    def serialize_data(self) -> str:
        if self.generate_message_id:
            self.set_anymail_id()

        headers = self.data["headers"]
        if self.is_batch():
            # Remove the all-recipient "to" header for batch sends.
            # Unisender Go will construct a single-recipient "to" for each recipient.
            # Unisender Go doesn't allow a "cc" header without an explicit "to"
            # header, so we cannot support "cc" for batch sends.
            headers.pop("to", None)
            if headers.pop("cc", None):
                self.unsupported_feature(
                    "cc with batch send (merge_data, merge_metadata, or merge_headers)"
                )

        if not headers:
            del self.data["headers"]  # don't send empty headers

        return self.serialize_json({"message": self.data})

    def set_anymail_id(self) -> None:
        """Ensure each personalization has a known anymail_id for event tracking"""
        for recipient in self.data["recipients"]:
            # This ensures duplicate recipients get same anymail_id
            # (because Unisender Go only sends to first instance of duplicate)
            email_address = recipient["email"]
            anymail_id = self.message_ids.get(email_address) or str(uuid.uuid4())
            recipient.setdefault("metadata", {})["anymail_id"] = anymail_id
            self.message_ids[email_address] = anymail_id

    #
    # Payload construction
    #

    def set_from_email(self, email: EmailAddress) -> None:
        self.data["from_email"] = email.addr_spec
        if email.display_name:
            self.data["from_name"] = email.display_name

    def _format_email_address(self, address):
        """
        Return EmailAddress address formatted for use with Unisender Go to/cc headers.

        Works around a bug in Unisender Go's API that rejects To or Cc headers
        containing commas, angle brackets, or @ in any display-name, despite those
        names being properly enclosed in "quotes" per RFC 5322. Workaround substitutes
        an RFC 2047 encoded word, which avoids the problem characters.

        Note that parens, quote chars, and other special characters appearing
        in "quoted strings" don't cause problems. (Unisender Go tech support
        has confirmed the problem is limited to , < > @.)

        This workaround is only necessary in the To and Cc headers. Unisender Go
        properly formats commas and other characters in `to_name` and `from_name`.
        (But see set_reply_to for a related issue.)
        """
        formatted = address.address
        if self.backend.workaround_display_name_bugs:
            # Workaround: force RFC-2047 QP encoded word for display_name if it has
            # prohibited chars (and isn't already encoded in the formatted address)
            display_name = address.display_name
            if re.search(r"[,<>@]", display_name) and display_name in formatted:
                formatted = str(
                    Address(
                        display_name=QP_CHARSET.header_encode(address.display_name),
                        addr_spec=address.addr_spec,
                    )
                )
        return formatted

    def set_recipients(self, recipient_type: str, emails: list[EmailAddress]):
        for email in emails:
            recipient = {"email": email.addr_spec}
            if email.display_name:
                recipient["substitutions"] = {"to_name": email.display_name}
            self.data["recipients"].append(recipient)

        if emails and recipient_type in {"to", "cc"}:
            # Add "to" or "cc" header listing all recipients of type.
            # See https://godocs.unisender.ru/cc-and-bcc.
            # (For batch sends, these will be adjusted later in self.serialize_data.)
            self.data["headers"][recipient_type] = ", ".join(
                self._format_email_address(email) for email in emails
            )

    def set_subject(self, subject: str) -> None:
        if subject:
            self.data["subject"] = subject

    def set_reply_to(self, emails: list[EmailAddress]) -> None:
        # Unisender Go only supports a single address in the reply_to API param.
        if len(emails) > 1:
            self.unsupported_feature("multiple reply_to addresses")
        if len(emails) > 0:
            reply_to = emails[0]
            self.data["reply_to"] = reply_to.addr_spec
            display_name = reply_to.display_name
            if display_name:
                if self.backend.workaround_display_name_bugs:
                    # Unisender Go doesn't properly "quote" (RFC 5322) a `reply_to_name`
                    # containing special characters (comma, parens, etc.), resulting
                    # in an invalid Reply-To header that can cause problems when the
                    # recipient tries to reply. (They *do* properly handle special chars
                    # in `to_name` and `from_name`; this only affects `reply_to_name`.)
                    if reply_to.address.startswith('"'):  # requires quoted syntax
                        # Workaround: force RFC-2047 encoded word
                        display_name = QP_CHARSET.header_encode(display_name)
                self.data["reply_to_name"] = display_name

    def set_extra_headers(self, headers: dict[str, str]) -> None:
        self.data["headers"].update(headers)

    def set_text_body(self, body: str) -> None:
        if body:
            self.data.setdefault("body", {})["plaintext"] = body

    def set_html_body(self, body: str) -> None:
        if body:
            self.data.setdefault("body", {})["html"] = body

    def add_alternative(self, content: str, mimetype: str):
        if mimetype.lower() == "text/x-amp-html":
            if "amp" in self.data.get("body", {}):
                self.unsupported_feature("multiple amp-html parts")
            self.data.setdefault("body", {})["amp"] = content
        else:
            super().add_alternative(content, mimetype)

    def add_attachment(self, attachment: Attachment) -> None:
        name = attachment.cid if attachment.inline else attachment.name
        att = {
            "content": attachment.b64content,
            "type": attachment.mimetype,
            "name": name or "",  # required - submit empty string if unknown
        }
        if attachment.inline:
            self.data.setdefault("inline_attachments", []).append(att)
        else:
            self.data.setdefault("attachments", []).append(att)

    def set_metadata(self, metadata: dict[str, str]) -> None:
        self.data["global_metadata"] = metadata

    def set_send_at(self, send_at: datetime | str) -> None:
        try:
            # "Date and time in the format “YYYY-MM-DD hh:mm:ss” in the UTC time zone."
            # If send_at is a datetime, it's guaranteed to be aware, but maybe not UTC.
            # Convert to UTC, then strip tzinfo to avoid isoformat "+00:00" at end.
            send_at_utc = send_at.astimezone(timezone.utc).replace(tzinfo=None)
            send_at_formatted = send_at_utc.isoformat(sep=" ", timespec="seconds")
            assert len(send_at_formatted) == 19
        except (AttributeError, TypeError):
            # Not a datetime - caller is responsible for formatting
            send_at_formatted = send_at
        self.data.setdefault("options", {})["send_at"] = send_at_formatted

    def set_tags(self, tags: list[str]) -> None:
        self.data["tags"] = tags

    def set_track_clicks(self, track_clicks: typing.Any):
        self.data["track_links"] = 1 if track_clicks else 0

    def set_track_opens(self, track_opens: typing.Any):
        self.data["track_read"] = 1 if track_opens else 0

    def set_template_id(self, template_id: str) -> None:
        self.data["template_id"] = template_id

    def set_merge_data(self, merge_data: dict[str, dict[str, str]]) -> None:
        if not merge_data:
            return
        assert self.data["recipients"]  # must be called after set_to
        for recipient in self.data["recipients"]:
            recipient_email = recipient["email"]
            if recipient_email in merge_data:
                # (substitutions may already be present with "to_email")
                recipient.setdefault("substitutions", {}).update(
                    merge_data[recipient_email]
                )

    def set_merge_global_data(self, merge_global_data: dict[str, str]) -> None:
        self.data["global_substitutions"] = merge_global_data

    def set_merge_metadata(self, merge_metadata: dict[str, str]) -> None:
        assert self.data["recipients"]  # must be called after set_to
        for recipient in self.data["recipients"]:
            recipient_email = recipient["email"]
            if recipient_email in merge_metadata:
                recipient["metadata"] = merge_metadata[recipient_email]

    # Unisender Go supports header substitution only with List-Unsubscribe.
    # (See https://godocs.unisender.ru/web-api-ref#email-send under "substitutions".)
    SUPPORTED_MERGE_HEADERS = {"List-Unsubscribe"}

    def set_merge_headers(self, merge_headers: dict[str, dict[str, str]]) -> None:
        assert self.data["recipients"]  # must be called after set_to
        if merge_headers:
            for recipient in self.data["recipients"]:
                recipient_email = recipient["email"]
                for key, value in merge_headers.get(recipient_email, {}).items():
                    field = key.title()  # canonicalize field name capitalization
                    if field in self.SUPPORTED_MERGE_HEADERS:
                        # Set up a substitution for Header__Field_Name
                        field_sub = "Header__" + field.replace("-", "_")
                        recipient.setdefault("substitutions", {})[field_sub] = value
                        self.data.setdefault("headers", {})[field] = (
                            "{{%s}}" % field_sub
                        )
                    else:
                        self.unsupported_feature(f"{field!r} in merge_headers")

    def set_esp_extra(self, extra: dict) -> None:
        update_deep(self.data, extra)



from django.conf import settings
from django.utils.encoding import force_str

from ..exceptions import AnymailRequestsAPIError
from ..message import AnymailRecipientStatus
from ..utils import get_anymail_setting, update_deep
from .base_requests import AnymailRequestsBackend, RequestsPayload


class EmailBackend(AnymailRequestsBackend):
    """
    SparkPost Email Backend
    """

    esp_name = "SparkPost"

    def __init__(self, **kwargs):
        """Init options from Django settings"""
        self.api_key = get_anymail_setting(
            "api_key", esp_name=self.esp_name, kwargs=kwargs, allow_bare=True
        )
        self.subaccount = get_anymail_setting(
            "subaccount", esp_name=self.esp_name, kwargs=kwargs, default=None
        )
        api_url = get_anymail_setting(
            "api_url",
            esp_name=self.esp_name,
            kwargs=kwargs,
            default="https://api.sparkpost.com/api/v1/",
        )
        if not api_url.endswith("/"):
            api_url += "/"
        super().__init__(api_url, **kwargs)

    def build_message_payload(self, message, defaults):
        return SparkPostPayload(message, defaults, self)

    def parse_recipient_status(self, response, payload, message):
        parsed_response = self.deserialize_json_response(response, payload, message)
        try:
            results = parsed_response["results"]
            accepted = results["total_accepted_recipients"]
            rejected = results["total_rejected_recipients"]
            transmission_id = results["id"]
        except (KeyError, TypeError) as err:
            raise AnymailRequestsAPIError(
                "Invalid SparkPost API response format",
                email_message=message,
                payload=payload,
                response=response,
                backend=self,
            ) from err

        # SparkPost doesn't (yet*) tell us *which* recipients were accepted or rejected.
        # (* looks like undocumented 'rcpt_to_errors' might provide this info.)
        # If all are one or the other, we can report a specific status;
        # else just report 'unknown' for all recipients.
        recipient_count = len(payload.recipients)
        if accepted == recipient_count and rejected == 0:
            status = "queued"
        elif rejected == recipient_count and accepted == 0:
            status = "rejected"
        else:  # mixed results, or wrong total
            status = "unknown"
        recipient_status = AnymailRecipientStatus(
            message_id=transmission_id, status=status
        )
        return {
            recipient.addr_spec: recipient_status for recipient in payload.recipients
        }


class SparkPostPayload(RequestsPayload):
    def __init__(self, message, defaults, backend, *args, **kwargs):
        http_headers = {
            "Authorization": backend.api_key,
            "Content-Type": "application/json",
        }
        if backend.subaccount is not None:
            http_headers["X-MSYS-SUBACCOUNT"] = backend.subaccount
        self.recipients = []  # all recipients, for backend parse_recipient_status
        self.cc_and_bcc = []  # for _finalize_recipients
        super().__init__(
            message, defaults, backend, headers=http_headers, *args, **kwargs
        )

    def get_api_endpoint(self):
        return "transmissions/"

    def serialize_data(self):
        self._finalize_recipients()
        self._check_content_options()
        return self.serialize_json(self.data)

    def _finalize_recipients(self):
        # https://www.sparkpost.com/docs/faq/cc-bcc-with-rest-api/
        # self.data["recipients"] is currently a list of all to-recipients. Must add all
        # cc and bcc recipients. Exactly how depends on whether this is a batch send.
        if self.is_batch():
            # For batch sends, must duplicate the cc/bcc for *every* to-recipient
            # (using each to-recipient's metadata and substitutions).
            extra_recipients = []
            for to_recipient in self.data["recipients"]:
                for email in self.cc_and_bcc:
                    extra = to_recipient.copy()  # gets "metadata" and "substitutions"
                    extra["address"] = {
                        "email": email.addr_spec,
                        "header_to": to_recipient["address"]["header_to"],
                    }
                    extra_recipients.append(extra)
            self.data["recipients"].extend(extra_recipients)
        else:
            # For non-batch sends, we need to patch up *everyone's* displayed
            # "To" header to show all the "To" recipients...
            full_to_header = ", ".join(
                to_recipient["address"]["header_to"]
                for to_recipient in self.data["recipients"]
            )
            for recipient in self.data["recipients"]:
                recipient["address"]["header_to"] = full_to_header
            # ... and then simply add the cc/bcc to the end of the list.
            # (There is no per-recipient data, or it would be a batch send.)
            self.data["recipients"].extend(
                {
                    "address": {
                        "email": email.addr_spec,
                        "header_to": full_to_header,
                    }
                }
                for email in self.cc_and_bcc
            )

    # SparkPost silently ignores certain "content" payload fields
    # when a template_id is used.
    IGNORED_WITH_TEMPLATE_ID = {
        # SparkPost API content.<field> -> feature name (for error message)
        "attachments": "attachments",
        "inline_images": "inline images",
        "headers": "extra headers and/or cc recipients",
        "from": "from_email",
        "reply_to": "reply_to",
    }

    def _check_content_options(self):
        if "template_id" in self.data["content"]:
            # subject, text, and html will cause 422 API Error:
            #     "message": "Both content object and template_id are specified",
            #     "code": "1301"
            # but others are silently ignored in a template send:
            ignored = [
                feature_name
                for field, feature_name in self.IGNORED_WITH_TEMPLATE_ID.items()
                if field in self.data["content"]
            ]
            if ignored:
                self.unsupported_feature("template_id with %s" % ", ".join(ignored))

    #
    # Payload construction
    #

    def init_payload(self):
        # The JSON payload:
        self.data = {
            "content": {},
            "recipients": [],
        }

    def set_from_email(self, email):
        if email:
            self.data["content"]["from"] = email.address

    def set_to(self, emails):
        if emails:
            # In the recipient address, "email" is the addr spec to deliver to,
            # and "header_to" is a fully-composed "To" header to display.
            # (We use "header_to" rather than "name" to simplify some logic
            # in _finalize_recipients; the results end up the same.)
            self.data["recipients"].extend(
                {
                    "address": {
                        "email": email.addr_spec,
                        "header_to": email.address,
                    }
                }
                for email in emails
            )
            self.recipients += emails

    def set_cc(self, emails):
        # https://www.sparkpost.com/docs/faq/cc-bcc-with-rest-api/
        if emails:
            # Add the Cc header, visible to all recipients:
            cc_header = ", ".join(email.address for email in emails)
            self.data["content"].setdefault("headers", {})["Cc"] = cc_header
            # Actual recipients are added later, in _finalize_recipients
            self.cc_and_bcc += emails
            self.recipients += emails

    def set_bcc(self, emails):
        if emails:
            # Actual recipients are added later, in _finalize_recipients
            self.cc_and_bcc += emails
            self.recipients += emails

    def set_subject(self, subject):
        self.data["content"]["subject"] = subject

    def set_reply_to(self, emails):
        if emails:
            self.data["content"]["reply_to"] = ", ".join(
                email.address for email in emails
            )

    def set_extra_headers(self, headers):
        if headers:
            self.data["content"].setdefault("headers", {}).update(headers)

    def set_text_body(self, body):
        self.data["content"]["text"] = body

    def set_html_body(self, body):
        if "html" in self.data["content"]:
            # second html body could show up through multiple alternatives,
            # or html body + alternative
            self.unsupported_feature("multiple html parts")
        self.data["content"]["html"] = body

    def add_alternative(self, content, mimetype):
        if mimetype.lower() == "text/x-amp-html":
            if "amp_html" in self.data["content"]:
                self.unsupported_feature("multiple html parts")
            self.data["content"]["amp_html"] = content
        else:
            super().add_alternative(content, mimetype)

    def set_attachments(self, atts):
        attachments = [
            {
                "name": att.name or "",
                "type": att.content_type,
                "data": att.b64content,
            }
            for att in atts
            if not att.inline
        ]
        if attachments:
            self.data["content"]["attachments"] = attachments

        inline_images = [
            {
                "name": att.cid,
                "type": att.mimetype,
                "data": att.b64content,
            }
            for att in atts
            if att.inline
        ]
        if inline_images:
            self.data["content"]["inline_images"] = inline_images

    # Anymail-specific payload construction
    def set_envelope_sender(self, email):
        self.data["return_path"] = email.addr_spec

    def set_metadata(self, metadata):
        self.data["metadata"] = metadata

    def set_merge_metadata(self, merge_metadata):
        for recipient in self.data["recipients"]:
            to_email = recipient["address"]["email"]
            if to_email in merge_metadata:
                recipient["metadata"] = merge_metadata[to_email]

    def set_merge_headers(self, merge_headers):
        def header_var(field):
            return "Header__" + field.title().replace("-", "_")

        merge_header_fields = set()

        for recipient in self.data["recipients"]:
            to_email = recipient["address"]["email"]
            if to_email in merge_headers:
                recipient_headers = merge_headers[to_email]
                recipient.setdefault("substitution_data", {}).update(
                    {header_var(key): value for key, value in recipient_headers.items()}
                )
                merge_header_fields.update(recipient_headers.keys())

        if merge_header_fields:
            headers = self.data.setdefault("content", {}).setdefault("headers", {})
            # Global substitution_data supplies defaults for defined headers:
            self.data.setdefault("substitution_data", {}).update(
                {
                    header_var(field): headers[field]
                    for field in merge_header_fields
                    if field in headers
                }
            )
            # Indirect merge_headers through substitution_data:
            headers.update(
                {field: "{{%s}}" % header_var(field) for field in merge_header_fields}
            )

    def set_send_at(self, send_at):
        try:
            start_time = send_at.replace(microsecond=0).isoformat()
        except (AttributeError, TypeError):
            start_time = send_at  # assume user already formatted
        self.data.setdefault("options", {})["start_time"] = start_time

    def set_tags(self, tags):
        if len(tags) > 0:
            self.data["campaign_id"] = tags[0]
            if len(tags) > 1:
                self.unsupported_feature("multiple tags (%r)" % tags)

    def set_track_clicks(self, track_clicks):
        self.data.setdefault("options", {})["click_tracking"] = track_clicks

    def set_track_opens(self, track_opens):
        self.data.setdefault("options", {})["open_tracking"] = track_opens

    def set_template_id(self, template_id):
        self.data["content"]["template_id"] = template_id
        # Must remove empty string "content" params when using stored template.
        # (Non-empty params are left in place, to cause API error.)
        for content_param in ["subject", "text", "html"]:
            try:
                if not self.data["content"][content_param]:
                    del self.data["content"][content_param]
            except KeyError:
                pass
        # "from" is also silently ignored. Strip it if empty or DEFAULT_FROM_EMAIL,
        # else leave in place to cause error in _check_content_options.
        try:
            from_email = self.data["content"]["from"]
            if not from_email or from_email == force_str(settings.DEFAULT_FROM_EMAIL):
                del self.data["content"]["from"]
        except KeyError:
            pass

    def set_merge_data(self, merge_data):
        for recipient in self.data["recipients"]:
            to_email = recipient["address"]["email"]
            if to_email in merge_data:
                recipient["substitution_data"] = merge_data[to_email]

    def set_merge_global_data(self, merge_global_data):
        self.data["substitution_data"] = merge_global_data

    # ESP-specific payload construction
    def set_esp_extra(self, extra):
        update_deep(self.data, extra)



#!/usr/bin/env python
# Render a README file (roughly) as it would appear on PyPI

import argparse
import sys
from importlib.metadata import PackageNotFoundError, metadata
from pathlib import Path
from typing import Dict, Optional

import readme_renderer.rst
from docutils.core import publish_string
from docutils.utils import SystemMessage

# Docutils template.txt in our directory:
DEFAULT_TEMPLATE_FILE = Path(__file__).with_name("template.txt").absolute()


def get_package_readme(package: str) -> str:
    # Note: "description" was added to metadata in Python 3.10
    return metadata(package)["description"]


class ReadMeHTMLWriter(readme_renderer.rst.Writer):
    translator_class = readme_renderer.rst.ReadMeHTMLTranslator

    def interpolation_dict(self) -> Dict[str, str]:
        result = super().interpolation_dict()
        # clean the same parts as readme_renderer.rst.render:
        clean = readme_renderer.rst.clean
        result["docinfo"] = clean(result["docinfo"])
        result["body"] = result["fragment"] = clean(result["fragment"])
        return result


def render(source_text: str, warning_stream=sys.stderr) -> Optional[str]:
    # Adapted from readme_renderer.rst.render
    settings = readme_renderer.rst.SETTINGS.copy()
    settings.update(
        {
            "warning_stream": warning_stream,
            "template": DEFAULT_TEMPLATE_FILE,
            # Input and output are text str (we handle decoding/encoding):
            "input_encoding": "unicode",
            "output_encoding": "unicode",
            # Exit with error on docutils warning or above.
            # (There's discussion of having readme_renderer ignore warnings;
            # this ensures they'll be treated as errors here.)
            "halt_level": 2,  # (docutils.utils.Reporter.WARNING_LEVEL)
            # Report all docutils warnings or above.
            # (The readme_renderer default suppresses this output.)
            "report_level": 2,  # (docutils.utils.Reporter.WARNING_LEVEL)
        }
    )

    writer = ReadMeHTMLWriter()

    try:
        return publish_string(
            source_text,
            writer=writer,
            settings_overrides=settings,
        )
    except SystemMessage:
        warning_stream.write("Error rendering readme source.\n")
        return None


def main(argv=None):
    parser = argparse.ArgumentParser(
        description="Render readme file as it would appear on PyPI"
    )
    input_group = parser.add_mutually_exclusive_group(required=True)
    input_group.add_argument(
        "-p", "--package", help="Source readme from package's metadata"
    )
    input_group.add_argument(
        "-i",
        "--input",
        help="Source readme.rst file ('-' for stdin)",
        type=argparse.FileType("r"),
    )
    parser.add_argument(
        "-o",
        "--output",
        help="Output file (default: stdout)",
        type=argparse.FileType("w"),
        default="-",
    )

    args = parser.parse_args(argv)
    if args.package:
        try:
            source_text = get_package_readme(args.package)
        except PackageNotFoundError:
            print(f"Package not installed: {args.package!r}", file=sys.stderr)
            sys.exit(2)
        if source_text is None:
            print(f"No metadata readme for {args.package!r}", file=sys.stderr)
            sys.exit(2)
    else:
        source_text = args.input.read()
    rendered = render(source_text)
    if rendered is None:
        sys.exit(2)
    args.output.write(rendered)


if __name__ == "__main__":
    main()



import os
import unittest
from email.utils import formataddr

from django.core import mail
from django.test import SimpleTestCase, override_settings, tag

from anymail.exceptions import AnymailAPIError, AnymailRecipientsRefused
from anymail.message import AnymailMessage

from .utils import AnymailTestMixin, sample_image_path

ANYMAIL_TEST_MANDRILL_API_KEY = os.getenv("ANYMAIL_TEST_MANDRILL_API_KEY")
ANYMAIL_TEST_MANDRILL_DOMAIN = os.getenv("ANYMAIL_TEST_MANDRILL_DOMAIN")


@tag("mandrill", "live")
@unittest.skipUnless(
    ANYMAIL_TEST_MANDRILL_API_KEY and ANYMAIL_TEST_MANDRILL_DOMAIN,
    "Set ANYMAIL_TEST_MANDRILL_API_KEY and ANYMAIL_TEST_MANDRILL_DOMAIN "
    "environment variables to run integration tests",
)
@override_settings(
    MANDRILL_API_KEY=ANYMAIL_TEST_MANDRILL_API_KEY,
    EMAIL_BACKEND="anymail.backends.mandrill.EmailBackend",
)
class MandrillBackendIntegrationTests(AnymailTestMixin, SimpleTestCase):
    """Mandrill API integration tests

    These tests run against the **live** Mandrill API, using the
    environment variable `ANYMAIL_TEST_MANDRILL_API_KEY` as the API key.
    If that variable is not set, these tests won't run.

    See https://mailchimp.com/developer/transactional/docs/fundamentals/#test-mode
    for info on Mandrill test keys.
    """

    def setUp(self):
        super().setUp()
        self.from_email = self.addr("from")
        self.message = mail.EmailMultiAlternatives(
            "Anymail Mandrill integration test",
            "Text content",
            self.from_email,
            [self.addr("test+to1")],
        )
        self.message.attach_alternative("<p>HTML content</p>", "text/html")

    def addr(self, username, display_name=None):
        """Construct test email address within our test domain"""
        # Because integration tests run within a Mandrill trial account,
        # both sender and recipient addresses must be within the test domain.
        # (Other recipient addresses will be rejected with 'recipient-domain-mismatch'.)
        email = "{username}@{domain}".format(
            username=username, domain=ANYMAIL_TEST_MANDRILL_DOMAIN
        )
        if display_name is not None:
            return formataddr((display_name, email))
        else:
            return email

    def test_simple_send(self):
        # Example of getting the Mandrill send status and _id from the message
        sent_count = self.message.send()
        self.assertEqual(sent_count, 1)

        # noinspection PyUnresolvedReferences
        anymail_status = self.message.anymail_status
        to_email = self.message.to[0]  # test+to1@{ANYMAIL_TEST_MANDRILL_DOMAIN}
        sent_status = anymail_status.recipients[to_email].status
        message_id = anymail_status.recipients[to_email].message_id

        # successful send (could still bounce later):
        self.assertIn(sent_status, ["sent", "queued"])
        # don't know what it'll be, but it should exist:
        self.assertGreater(len(message_id), 0)

        # set of all recipient statuses:
        self.assertEqual(anymail_status.status, {sent_status})
        # because only a single recipient (else would be a set):
        self.assertEqual(anymail_status.message_id, message_id)

    def test_all_options(self):
        message = AnymailMessage(
            subject="Anymail Mandrill all-options integration test",
            body="This is the text body",
            from_email=self.addr("from", "Test From, with comma"),
            to=[self.addr("test+to1"), self.addr("test+to2", "Recipient 2")],
            cc=[self.addr("test+cc1"), self.addr("test+cc2", "Copy 2")],
            bcc=[self.addr("test+bcc1"), self.addr("test+bcc2", "Blind Copy 2")],
            reply_to=["reply1@example.com", "Reply 2 <reply2@example.com>"],
            headers={"X-Anymail-Test": "value"},
            # no metadata, send_at, track_clicks support
            tags=["tag 1"],  # max one tag
            track_opens=True,
        )
        message.attach("attachment1.txt", "Here is some\ntext for you", "text/plain")
        message.attach("attachment2.csv", "ID,Name\n1,Amy Lina", "text/csv")
        cid = message.attach_inline_image_file(sample_image_path())
        message.attach_alternative(
            "<p><b>HTML:</b> with <a href='http://example.com'>link</a>"
            "and image: <img src='cid:%s'></div>" % cid,
            "text/html",
        )

        message.send()
        self.assertTrue(message.anymail_status.status.issubset({"queued", "sent"}))

    def test_invalid_from(self):
        # Example of trying to send from an invalid address
        self.message.from_email = (
            "webmaster@localhost"  # Django default DEFAULT_FROM_EMAIL
        )
        with self.assertRaisesMessage(AnymailAPIError, "email address is invalid"):
            self.message.send()

    def test_invalid_to(self):
        # Example of detecting when a recipient is not a valid email address
        self.message.to = ["invalid@localhost"]
        try:
            self.message.send()
        except AnymailRecipientsRefused:
            # Mandrill refused to deliver the mail -- message.anymail_status
            # will tell you why:
            # noinspection PyUnresolvedReferences
            anymail_status = self.message.anymail_status
            self.assertEqual(
                anymail_status.recipients["invalid@localhost"].status, "invalid"
            )
            self.assertEqual(anymail_status.status, {"invalid"})
        else:
            # Sometimes Mandrill queues these test sends
            # noinspection PyUnresolvedReferences
            if self.message.anymail_status.status == {"queued"}:
                self.skipTest("Mandrill queued the send -- can't complete this test")
            else:
                self.fail(
                    "Anymail did not raise AnymailRecipientsRefused"
                    " for invalid recipient"
                )

    def test_rejected_to(self):
        # Example of detecting when a recipient is on Mandrill's rejection blacklist
        self.message.to = ["reject@test.mandrillapp.com"]
        try:
            self.message.send()
        except AnymailRecipientsRefused:
            # Mandrill refused to deliver the mail -- message.anymail_status will
            # tell you why:
            # noinspection PyUnresolvedReferences
            anymail_status = self.message.anymail_status
            self.assertEqual(
                anymail_status.recipients["reject@test.mandrillapp.com"].status,
                "rejected",
            )
            self.assertEqual(anymail_status.status, {"rejected"})
        else:
            # Sometimes Mandrill queues these test sends
            # noinspection PyUnresolvedReferences
            if self.message.anymail_status.status == {"queued"}:
                self.skipTest("Mandrill queued the send -- can't complete this test")
            else:
                self.fail(
                    "Anymail did not raise AnymailRecipientsRefused"
                    " for blacklist recipient"
                )

    @override_settings(MANDRILL_API_KEY="Hey, that's not an API key!")
    def test_invalid_api_key(self):
        # Example of trying to send with an invalid MANDRILL_API_KEY.
        # Mandrill responds either 401 or 500 status for this case (it varies),
        # but always seems to include the message "Invalid API key".
        # Either response should result in an AnymailAPIError.
        with self.assertRaisesMessage(AnymailAPIError, "Invalid API key"):
            self.message.send()



import os
import unittest
from email.utils import formataddr

from django.test import SimpleTestCase, override_settings, tag

from anymail.exceptions import AnymailAPIError
from anymail.message import AnymailMessage

from .utils import AnymailTestMixin, sample_image_content

ANYMAIL_TEST_MAILJET_API_KEY = os.getenv("ANYMAIL_TEST_MAILJET_API_KEY")
ANYMAIL_TEST_MAILJET_SECRET_KEY = os.getenv("ANYMAIL_TEST_MAILJET_SECRET_KEY")
ANYMAIL_TEST_MAILJET_DOMAIN = os.getenv("ANYMAIL_TEST_MAILJET_DOMAIN")
ANYMAIL_TEST_MAILJET_TEMPLATE_ID = os.getenv("ANYMAIL_TEST_MAILJET_TEMPLATE_ID")


@tag("mailjet", "live")
@unittest.skipUnless(
    ANYMAIL_TEST_MAILJET_API_KEY
    and ANYMAIL_TEST_MAILJET_SECRET_KEY
    and ANYMAIL_TEST_MAILJET_DOMAIN,
    "Set ANYMAIL_TEST_MAILJET_API_KEY and ANYMAIL_TEST_MAILJET_SECRET_KEY"
    " and ANYMAIL_TEST_MAILJET_DOMAIN environment variables to run Mailjet"
    " integration tests",
)
@override_settings(
    ANYMAIL={
        "MAILJET_API_KEY": ANYMAIL_TEST_MAILJET_API_KEY,
        "MAILJET_SECRET_KEY": ANYMAIL_TEST_MAILJET_SECRET_KEY,
        "MAILJET_SEND_DEFAULTS": {
            "esp_extra": {"SandboxMode": True}  # don't actually send mail
        },
    },
    EMAIL_BACKEND="anymail.backends.mailjet.EmailBackend",
)
class MailjetBackendIntegrationTests(AnymailTestMixin, SimpleTestCase):
    """
    Mailjet API integration tests

    These tests run against the **live** Mailjet API, using the environment variables
    `ANYMAIL_TEST_MAILJET_API_KEY` and `ANYMAIL_TEST_MAILJET_SECRET_KEY` as the API key
    and API secret key, respectively, and `ANYMAIL_TEST_MAILJET_DOMAIN` as a validated
    Mailjet sending domain. If those variables are not set, these tests won't run.

    These tests enable Mailjet's SandboxMode to avoid sending any email;
    remove the esp_extra setting above if you are trying to actually send test messages.
    """

    def setUp(self):
        super().setUp()
        self.from_email = "test@%s" % ANYMAIL_TEST_MAILJET_DOMAIN
        self.message = AnymailMessage(
            "Anymail Mailjet integration test",
            "Text content",
            self.from_email,
            ["test+to1@anymail.dev"],
        )
        self.message.attach_alternative("<p>HTML content</p>", "text/html")

    def test_simple_send(self):
        # Example of getting the Mailjet send status and message id from the message
        sent_count = self.message.send()
        self.assertEqual(sent_count, 1)

        anymail_status = self.message.anymail_status
        sent_status = anymail_status.recipients["test+to1@anymail.dev"].status
        message_id = anymail_status.recipients["test+to1@anymail.dev"].message_id

        self.assertEqual(sent_status, "sent")
        self.assertRegex(message_id, r".+")
        # set of all recipient statuses:
        self.assertEqual(anymail_status.status, {sent_status})
        self.assertEqual(anymail_status.message_id, message_id)

    def test_all_options(self):
        message = AnymailMessage(
            subject="Anymail Mailjet all-options integration test",
            body="This is the text body",
            from_email=formataddr(("Test Sender, Inc.", self.from_email)),
            to=["test+to1@anymail.dev", '"Recipient, 2nd" <test+to2@anymail.dev>'],
            cc=["test+cc1@anymail.dev", "Copy 2 <test+cc1@anymail.dev>"],
            bcc=["test+bcc1@anymail.dev", "Blind Copy 2 <test+bcc2@anymail.dev>"],
            # Mailjet only supports single reply_to:
            reply_to=['"Reply, To" <reply2@example.com>'],
            headers={"X-Anymail-Test": "value"},
            metadata={"meta1": "simple string", "meta2": 2},
            tags=["tag 1"],  # Mailjet only allows a single tag
            track_clicks=True,
            track_opens=True,
        )
        message.attach("attachment1.txt", "Here is some\ntext for you", "text/plain")
        message.attach("attachment2.csv", "ID,Name\n1,Amy Lina", "text/csv")
        cid = message.attach_inline_image(sample_image_content())
        message.attach_alternative(
            "<p><b>HTML:</b> with <a href='http://example.com'>link</a>"
            "and image: <img src='cid:%s'></div>" % cid,
            "text/html",
        )

        message.send()
        self.assertEqual(message.anymail_status.status, {"sent"})

    def test_merge_data(self):
        message = AnymailMessage(
            # Mailjet doesn't support merge fields in the subject
            subject="Anymail Mailjet merge_data test",
            body="This body includes merge data: [[var:value]]\n"
            "And global merge data: [[var:global]]",
            from_email=formataddr(("Test From", self.from_email)),
            to=["test+to1@anymail.dev", "Recipient 2 <test+to2@anymail.dev>"],
            merge_data={
                "test+to1@anymail.dev": {"value": "one"},
                "test+to2@anymail.dev": {"value": "two"},
            },
            merge_global_data={"global": "global_value"},
            metadata={"customer-id": "unknown", "meta2": 2},
            merge_metadata={
                "test+to1@anymail.dev": {"customer-id": "ZXK9123"},
                "test+to2@anymail.dev": {"customer-id": "ZZT4192"},
            },
            headers={
                "List-Unsubscribe-Post": "List-Unsubscribe=One-Click",
                "List-Unsubscribe": "<mailto:unsubscribe@example.com>",
            },
            merge_headers={
                "test+to1@anymail.dev": {
                    "List-Unsubscribe": "<https://example.com/a/>",
                },
                "test+to2@anymail.dev": {
                    "List-Unsubscribe": "<https://example.com/b/>",
                },
            },
        )
        message.send()
        recipient_status = message.anymail_status.recipients
        self.assertEqual(recipient_status["test+to1@anymail.dev"].status, "sent")
        self.assertEqual(recipient_status["test+to2@anymail.dev"].status, "sent")

    @unittest.skipUnless(
        ANYMAIL_TEST_MAILJET_TEMPLATE_ID,
        "Set ANYMAIL_TEST_MAILJET_TEMPLATE_ID environment variable to run"
        " Mailjet stored template integration test",
    )
    def test_stored_template(self):
        message = AnymailMessage(
            # ID of the real template named 'test-template' in our Mailjet test account:
            template_id=ANYMAIL_TEST_MAILJET_TEMPLATE_ID,
            to=["test+to1@anymail.dev"],
            merge_data={
                "test+to1@anymail.dev": {
                    "name": "Test Recipient",
                }
            },
            merge_global_data={
                "order": "12345",
            },
        )
        message.from_email = None  # use the template's sender email/name
        message.send()
        recipient_status = message.anymail_status.recipients
        self.assertEqual(recipient_status["test+to1@anymail.dev"].status, "sent")

    @override_settings(
        ANYMAIL={
            "MAILJET_API_KEY": "Hey, that's not an API key!",
            "MAILJET_SECRET_KEY": "and this isn't the secret for it",
        }
    )
    def test_invalid_api_key(self):
        with self.assertRaises(AnymailAPIError) as cm:
            self.message.send()
        err = cm.exception
        self.assertEqual(err.status_code, 401)
        self.assertIn("API key authentication/authorization failure", str(err))



import json
from email.utils import unquote

from django.utils.dateparse import parse_datetime

from ..exceptions import AnymailConfigurationError
from ..inbound import AnymailInboundMessage
from ..signals import (
    AnymailInboundEvent,
    AnymailTrackingEvent,
    EventType,
    RejectReason,
    inbound,
    tracking,
)
from ..utils import EmailAddress, getfirst
from .base import AnymailBaseWebhookView


class PostmarkBaseWebhookView(AnymailBaseWebhookView):
    """Base view class for Postmark webhooks"""

    esp_name = "Postmark"

    def parse_events(self, request):
        esp_event = json.loads(request.body.decode("utf-8"))
        return [self.esp_to_anymail_event(esp_event)]

    def esp_to_anymail_event(self, esp_event):
        raise NotImplementedError()


class PostmarkTrackingWebhookView(PostmarkBaseWebhookView):
    """Handler for Postmark delivery and engagement tracking webhooks"""

    signal = tracking

    event_record_types = {
        # Map Postmark event RecordType --> Anymail normalized event type
        "Bounce": EventType.BOUNCED,  # but check Type field for further info (below)
        "Click": EventType.CLICKED,
        "Delivery": EventType.DELIVERED,
        "Open": EventType.OPENED,
        "SpamComplaint": EventType.COMPLAINED,
        "SubscriptionChange": EventType.UNSUBSCRIBED,
        "Inbound": EventType.INBOUND,  # future, probably
    }

    event_types = {
        # Map Postmark bounce/spam event Type
        #   --> Anymail normalized (event type, reject reason)
        "HardBounce": (EventType.BOUNCED, RejectReason.BOUNCED),
        "Transient": (EventType.DEFERRED, None),
        "Unsubscribe": (EventType.UNSUBSCRIBED, RejectReason.UNSUBSCRIBED),
        "Subscribe": (EventType.SUBSCRIBED, None),
        "AutoResponder": (EventType.AUTORESPONDED, None),
        "AddressChange": (EventType.AUTORESPONDED, None),
        "DnsError": (EventType.DEFERRED, None),  # "temporary DNS error"
        "SpamNotification": (EventType.COMPLAINED, RejectReason.SPAM),
        # Receiving MTA is testing Postmark:
        "OpenRelayTest": (EventType.DEFERRED, None),
        "Unknown": (EventType.UNKNOWN, None),
        # might also receive HardBounce later:
        "SoftBounce": (EventType.BOUNCED, RejectReason.BOUNCED),
        "VirusNotification": (EventType.BOUNCED, RejectReason.OTHER),
        "ChallengeVerification": (EventType.AUTORESPONDED, None),
        "BadEmailAddress": (EventType.REJECTED, RejectReason.INVALID),
        "SpamComplaint": (EventType.COMPLAINED, RejectReason.SPAM),
        "ManuallyDeactivated": (EventType.REJECTED, RejectReason.BLOCKED),
        "Unconfirmed": (EventType.REJECTED, None),
        "Blocked": (EventType.REJECTED, RejectReason.BLOCKED),
        # could occur if user also using Postmark SMTP directly:
        "SMTPApiError": (EventType.FAILED, None),
        "InboundError": (EventType.INBOUND_FAILED, None),
        "DMARCPolicy": (EventType.REJECTED, RejectReason.BLOCKED),
        "TemplateRenderingFailed": (EventType.FAILED, None),
        "ManualSuppression": (EventType.UNSUBSCRIBED, RejectReason.UNSUBSCRIBED),
    }

    def esp_to_anymail_event(self, esp_event):
        reject_reason = None
        try:
            esp_record_type = esp_event["RecordType"]
        except KeyError:
            if "FromFull" in esp_event:
                # This is an inbound event
                event_type = EventType.INBOUND
            else:
                event_type = EventType.UNKNOWN
        else:
            event_type = self.event_record_types.get(esp_record_type, EventType.UNKNOWN)

        if event_type == EventType.INBOUND:
            raise AnymailConfigurationError(
                "You seem to have set Postmark's *inbound* webhook "
                "to Anymail's Postmark *tracking* webhook URL."
            )

        if event_type in (EventType.BOUNCED, EventType.COMPLAINED):
            # additional info is in the Type field
            try:
                event_type, reject_reason = self.event_types[esp_event["Type"]]
            except KeyError:
                pass
        if event_type == EventType.UNSUBSCRIBED:
            if esp_event["SuppressSending"]:
                # Postmark doesn't provide a way to distinguish between
                # explicit unsubscribes and bounces
                try:
                    event_type, reject_reason = self.event_types[
                        esp_event["SuppressionReason"]
                    ]
                except KeyError:
                    pass
            else:
                event_type, reject_reason = self.event_types["Subscribe"]

        # Email for bounce; Recipient for open:
        recipient = getfirst(esp_event, ["Email", "Recipient"], None)

        try:
            timestr = getfirst(
                esp_event, ["DeliveredAt", "BouncedAt", "ReceivedAt", "ChangedAt"]
            )
        except KeyError:
            timestamp = None
        else:
            timestamp = parse_datetime(timestr)

        try:
            event_id = str(esp_event["ID"])  # only in bounce events
        except KeyError:
            event_id = None

        metadata = esp_event.get("Metadata", {})
        try:
            tags = [esp_event["Tag"]]
        except KeyError:
            tags = []

        return AnymailTrackingEvent(
            description=esp_event.get("Description", None),
            esp_event=esp_event,
            event_id=event_id,
            event_type=event_type,
            message_id=esp_event.get("MessageID", None),
            metadata=metadata,
            mta_response=esp_event.get("Details", None),
            recipient=recipient,
            reject_reason=reject_reason,
            tags=tags,
            timestamp=timestamp,
            user_agent=esp_event.get("UserAgent", None),
            click_url=esp_event.get("OriginalLink", None),
        )


class PostmarkInboundWebhookView(PostmarkBaseWebhookView):
    """Handler for Postmark inbound webhook"""

    signal = inbound

    def esp_to_anymail_event(self, esp_event):
        # Check correct webhook (inbound events don't have RecordType):
        esp_record_type = esp_event.get("RecordType", "Inbound")
        if esp_record_type != "Inbound":
            raise AnymailConfigurationError(
                f"You seem to have set Postmark's *{esp_record_type}* webhook"
                f" to Anymail's Postmark *inbound* webhook URL."
            )

        headers = esp_event.get("Headers", [])

        # Postmark inbound prepends "Return-Path" to Headers list
        # (but it doesn't appear in original message or RawEmail).
        # (A Return-Path anywhere else in the headers or RawEmail
        # can't be considered legitimate.)
        envelope_sender = None
        if len(headers) > 0 and headers[0]["Name"].lower() == "return-path":
            envelope_sender = unquote(headers[0]["Value"])  # remove <>
            headers = headers[1:]  # don't include in message construction

        if "RawEmail" in esp_event:
            message = AnymailInboundMessage.parse_raw_mime(esp_event["RawEmail"])
            # Postmark provides Bcc when delivered-to is not in To header,
            # but doesn't add it to the RawEmail.
            if esp_event.get("BccFull") and "Bcc" not in message:
                message["Bcc"] = self._addresses(esp_event["BccFull"])

        else:
            # RawEmail not included in payload; construct from parsed data.
            attachments = [
                AnymailInboundMessage.construct_attachment(
                    content_type=attachment["ContentType"],
                    # Real payloads have "Content", test payloads have "Data" (?!):
                    content=attachment.get("Content") or attachment["Data"],
                    base64=True,
                    filename=attachment.get("Name"),
                    content_id=attachment.get("ContentID"),
                )
                for attachment in esp_event.get("Attachments", [])
            ]
            message = AnymailInboundMessage.construct(
                from_email=self._address(esp_event.get("FromFull")),
                to=self._addresses(esp_event.get("ToFull")),
                cc=self._addresses(esp_event.get("CcFull")),
                bcc=self._addresses(esp_event.get("BccFull")),
                subject=esp_event.get("Subject", ""),
                headers=((header["Name"], header["Value"]) for header in headers),
                text=esp_event.get("TextBody", ""),
                html=esp_event.get("HtmlBody", ""),
                attachments=attachments,
            )
            # Postmark strips these headers and provides them as separate event fields:
            if esp_event.get("Date") and "Date" not in message:
                message["Date"] = esp_event["Date"]
            if esp_event.get("ReplyTo") and "Reply-To" not in message:
                message["Reply-To"] = esp_event["ReplyTo"]

        message.envelope_sender = envelope_sender
        message.envelope_recipient = esp_event.get("OriginalRecipient")
        message.stripped_text = esp_event.get("StrippedTextReply")

        message.spam_detected = message.get("X-Spam-Status", "No").lower() == "yes"
        try:
            message.spam_score = float(message["X-Spam-Score"])
        except (TypeError, ValueError):
            pass

        return AnymailInboundEvent(
            event_type=EventType.INBOUND,
            # Postmark doesn't provide inbound event timestamp:
            timestamp=None,
            # Postmark uuid, different from Message-ID mime header:
            event_id=esp_event.get("MessageID", None),
            esp_event=esp_event,
            message=message,
        )

    @classmethod
    def _address(cls, full):
        """
        Return a formatted email address
        from a Postmark inbound {From,To,Cc,Bcc}Full dict
        """
        if full is None:
            return ""
        return str(
            EmailAddress(
                display_name=full.get("Name", ""),
                addr_spec=full.get("Email", ""),
            )
        )

    @classmethod
    def _addresses(cls, full_list):
        """
        Return a formatted email address list string
        from a Postmark inbound {To,Cc,Bcc}Full[] list of dicts
        """
        if full_list is None:
            return None
        return ", ".join(cls._address(addr) for addr in full_list)



from django.dispatch import receiver

from anymail.backends.test import EmailBackend as TestEmailBackend
from anymail.exceptions import AnymailCancelSend, AnymailRecipientsRefused
from anymail.message import AnymailRecipientStatus
from anymail.signals import post_send, pre_send

from .test_general_backend import TestBackendTestCase


class TestPreSendSignal(TestBackendTestCase):
    """Test Anymail's pre_send signal"""

    def test_pre_send(self):
        """Pre-send receivers invoked for each message, before sending"""

        @receiver(pre_send, weak=False)
        def handle_pre_send(sender, message, esp_name, **kwargs):
            self.assertEqual(self.get_send_count(), 0)  # not sent yet
            self.assertEqual(sender, TestEmailBackend)
            self.assertEqual(message, self.message)
            # the TestEmailBackend's ESP is named "Test":
            self.assertEqual(esp_name, "Test")
            self.receiver_called = True

        self.addCleanup(pre_send.disconnect, receiver=handle_pre_send)

        self.receiver_called = False
        self.message.send()
        self.assertTrue(self.receiver_called)
        self.assertEqual(self.get_send_count(), 1)  # sent now

    def test_modify_message_in_pre_send(self):
        """Pre-send receivers can modify message"""

        @receiver(pre_send, weak=False)
        def handle_pre_send(sender, message, esp_name, **kwargs):
            message.to = [email for email in message.to if not email.startswith("bad")]
            message.body += "\nIf you have received this message in error, ignore it"

        self.addCleanup(pre_send.disconnect, receiver=handle_pre_send)

        self.message.to = ["legit@example.com", "bad@example.com"]
        self.message.send()
        params = self.get_send_params()
        self.assertEqual(
            # params['to'] is EmailAddress list:
            [email.addr_spec for email in params["to"]],
            ["legit@example.com"],
        )
        self.assertRegex(
            params["text_body"],
            r"If you have received this message in error, ignore it$",
        )

    def test_cancel_in_pre_send(self):
        """Pre-send receiver can cancel the send"""

        @receiver(pre_send, weak=False)
        def cancel_pre_send(sender, message, esp_name, **kwargs):
            raise AnymailCancelSend("whoa there")

        self.addCleanup(pre_send.disconnect, receiver=cancel_pre_send)

        self.message.send()
        self.assertEqual(self.get_send_count(), 0)  # send API not called


class TestPostSendSignal(TestBackendTestCase):
    """Test Anymail's post_send signal"""

    def test_post_send(self):
        """Post-send receiver called for each message, after sending"""

        @receiver(post_send, weak=False)
        def handle_post_send(sender, message, status, esp_name, **kwargs):
            self.assertEqual(self.get_send_count(), 1)  # already sent
            self.assertEqual(sender, TestEmailBackend)
            self.assertEqual(message, self.message)
            self.assertEqual(status.status, {"sent"})
            self.assertEqual(status.message_id, 0)
            self.assertEqual(status.recipients["to@example.com"].status, "sent")
            self.assertEqual(status.recipients["to@example.com"].message_id, 0)
            # the TestEmailBackend's ESP is named "Test":
            self.assertEqual(esp_name, "Test")
            self.receiver_called = True

        self.addCleanup(post_send.disconnect, receiver=handle_post_send)

        self.receiver_called = False
        self.message.send()
        self.assertTrue(self.receiver_called)

    def test_post_send_exception(self):
        """All post-send receivers called, even if one throws"""

        @receiver(post_send, weak=False)
        def handler_1(sender, message, status, esp_name, **kwargs):
            raise ValueError("oops")

        self.addCleanup(post_send.disconnect, receiver=handler_1)

        @receiver(post_send, weak=False)
        def handler_2(sender, message, status, esp_name, **kwargs):
            self.handler_2_called = True

        self.addCleanup(post_send.disconnect, receiver=handler_2)

        self.handler_2_called = False
        with self.assertRaises(ValueError):
            self.message.send()
        self.assertTrue(self.handler_2_called)

    def test_rejected_recipients(self):
        """Post-send receiver even if AnymailRecipientsRefused is raised"""

        @receiver(post_send, weak=False)
        def handle_post_send(sender, message, status, esp_name, **kwargs):
            self.receiver_called = True

        self.addCleanup(post_send.disconnect, receiver=handle_post_send)

        self.message.anymail_test_response = {
            "recipient_status": {
                "to@example.com": AnymailRecipientStatus(
                    message_id=None, status="rejected"
                )
            }
        }

        self.receiver_called = False
        with self.assertRaises(AnymailRecipientsRefused):
            self.message.send()
        self.assertTrue(self.receiver_called)



import json
from base64 import b64encode
from decimal import Decimal
from email.mime.base import MIMEBase
from email.mime.image import MIMEImage

from django.core import mail
from django.core.exceptions import ImproperlyConfigured
from django.test import SimpleTestCase, override_settings, tag

from anymail.exceptions import (
    AnymailAPIError,
    AnymailInvalidAddress,
    AnymailRecipientsRefused,
    AnymailSerializationError,
    AnymailUnsupportedFeature,
)
from anymail.message import AnymailMessage, attach_inline_image_file

from .mock_requests_backend import (
    RequestsBackendMockAPITestCase,
    SessionSharingTestCases,
)
from .utils import (
    SAMPLE_IMAGE_FILENAME,
    AnymailTestMixin,
    decode_att,
    sample_image_content,
    sample_image_path,
)


@tag("postmark")
@override_settings(
    EMAIL_BACKEND="anymail.backends.postmark.EmailBackend",
    ANYMAIL={"POSTMARK_SERVER_TOKEN": "test_server_token"},
)
class PostmarkBackendMockAPITestCase(RequestsBackendMockAPITestCase):
    DEFAULT_RAW_RESPONSE = b"""{
        "To": "to@example.com",
        "SubmittedAt": "2016-03-12T15:27:50.4468803-05:00",
        "MessageID": "b4007d94-33f1-4e78-a783-97417d6c80e6",
        "ErrorCode":0,
        "Message":"OK"
    }"""

    def setUp(self):
        super().setUp()
        # Simple message useful for many tests
        self.message = mail.EmailMultiAlternatives(
            "Subject", "Text Body", "from@example.com", ["to@example.com"]
        )


@tag("postmark")
class PostmarkBackendStandardEmailTests(PostmarkBackendMockAPITestCase):
    """Test backend support for Django standard email features"""

    def test_send_mail(self):
        """Test basic API for simple send"""
        mail.send_mail(
            "Subject here",
            "Here is the message.",
            "from@sender.example.com",
            ["to@example.com"],
            fail_silently=False,
        )
        self.assert_esp_called("/email")
        headers = self.get_api_call_headers()
        self.assertEqual(headers["X-Postmark-Server-Token"], "test_server_token")
        data = self.get_api_call_json()
        self.assertEqual(data["Subject"], "Subject here")
        self.assertEqual(data["TextBody"], "Here is the message.")
        self.assertEqual(data["From"], "from@sender.example.com")
        self.assertEqual(data["To"], "to@example.com")

    def test_name_addr(self):
        """Make sure RFC2822 name-addr format (with display-name) is allowed

        (Test both sender and recipient addresses)
        """
        msg = mail.EmailMessage(
            "Subject",
            "Message",
            "From Name <from@example.com>",
            ["Recipient #1 <to1@example.com>", "to2@example.com"],
            cc=["Carbon Copy <cc1@example.com>", "cc2@example.com"],
            bcc=["Blind Copy <bcc1@example.com>", "bcc2@example.com"],
        )
        msg.send()
        data = self.get_api_call_json()
        self.assertEqual(data["From"], "From Name <from@example.com>")
        self.assertEqual(data["To"], "Recipient #1 <to1@example.com>, to2@example.com")
        self.assertEqual(data["Cc"], "Carbon Copy <cc1@example.com>, cc2@example.com")
        self.assertEqual(data["Bcc"], "Blind Copy <bcc1@example.com>, bcc2@example.com")

    def test_email_message(self):
        email = mail.EmailMessage(
            "Subject",
            "Body goes here",
            "from@example.com",
            ["to1@example.com", "Also To <to2@example.com>"],
            bcc=["bcc1@example.com", "Also BCC <bcc2@example.com>"],
            cc=["cc1@example.com", "Also CC <cc2@example.com>"],
            headers={
                "Reply-To": "another@example.com",
                "X-MyHeader": "my value",
                # should override backend msgid:
                "Message-ID": "mycustommsgid@sales.example.com",
            },
        )
        email.send()
        data = self.get_api_call_json()
        self.assertEqual(data["Subject"], "Subject")
        self.assertEqual(data["TextBody"], "Body goes here")
        self.assertEqual(data["From"], "from@example.com")
        self.assertEqual(data["To"], "to1@example.com, Also To <to2@example.com>")
        self.assertEqual(data["Bcc"], "bcc1@example.com, Also BCC <bcc2@example.com>")
        self.assertEqual(data["Cc"], "cc1@example.com, Also CC <cc2@example.com>")
        self.assertEqual(data["ReplyTo"], "another@example.com")
        self.assertCountEqual(
            data["Headers"],
            [
                {"Name": "Message-ID", "Value": "mycustommsgid@sales.example.com"},
                {"Name": "X-MyHeader", "Value": "my value"},
            ],
        )

    def test_html_message(self):
        text_content = "This is an important message."
        html_content = "<p>This is an <strong>important</strong> message.</p>"
        email = mail.EmailMultiAlternatives(
            "Subject", text_content, "from@example.com", ["to@example.com"]
        )
        email.attach_alternative(html_content, "text/html")
        email.send()
        data = self.get_api_call_json()
        self.assertEqual(data["TextBody"], text_content)
        self.assertEqual(data["HtmlBody"], html_content)
        # Don't accidentally send the html part as an attachment:
        self.assertNotIn("Attachments", data)

    def test_html_only_message(self):
        html_content = "<p>This is an <strong>important</strong> message.</p>"
        email = mail.EmailMessage(
            "Subject", html_content, "from@example.com", ["to@example.com"]
        )
        email.content_subtype = "html"  # Main content is now text/html
        email.send()
        data = self.get_api_call_json()
        self.assertNotIn("TextBody", data)
        self.assertEqual(data["HtmlBody"], html_content)

    def test_extra_headers(self):
        self.message.extra_headers = {"X-Custom": "string", "X-Num": 123}
        self.message.send()
        data = self.get_api_call_json()
        self.assertCountEqual(
            data["Headers"],
            [{"Name": "X-Custom", "Value": "string"}, {"Name": "X-Num", "Value": 123}],
        )

    def test_extra_headers_serialization_error(self):
        self.message.extra_headers = {"X-Custom": Decimal(12.5)}
        with self.assertRaisesMessage(AnymailSerializationError, "Decimal"):
            self.message.send()

    def test_reply_to(self):
        email = mail.EmailMessage(
            "Subject",
            "Body goes here",
            "from@example.com",
            ["to1@example.com"],
            reply_to=["reply@example.com", "Other <reply2@example.com>"],
            headers={"X-Other": "Keep"},
        )
        email.send()
        data = self.get_api_call_json()
        self.assertEqual(
            data["ReplyTo"], "reply@example.com, Other <reply2@example.com>"
        )
        # don't lose other headers:
        self.assertEqual(data["Headers"], [{"Name": "X-Other", "Value": "Keep"}])

    def test_reply_to_header(self):
        # Reply-To needs to be moved out of headers, into dedicated param
        email = mail.EmailMessage(
            "Subject",
            "Body goes here",
            "from@example.com",
            ["to1@example.com"],
            headers={
                "reply-to": "reply@example.com, Other <reply2@example.com>",
                "X-Other": "Keep",
            },
        )
        email.send()
        data = self.get_api_call_json()
        self.assertEqual(
            data["ReplyTo"], "reply@example.com, Other <reply2@example.com>"
        )
        # don't lose other headers:
        self.assertEqual(data["Headers"], [{"Name": "X-Other", "Value": "Keep"}])

    def test_attachments(self):
        text_content = "* Item one\n* Item two\n* Item three"
        self.message.attach(
            filename="test.txt", content=text_content, mimetype="text/plain"
        )

        # Should guess mimetype if not provided...
        png_content = b"PNG\xb4 pretend this is the contents of a png file"
        self.message.attach(filename="test.png", content=png_content)

        # Should work with a MIMEBase object (also tests no filename)...
        pdf_content = b"PDF\xb4 pretend this is valid pdf data"
        mimeattachment = MIMEBase("application", "pdf")
        mimeattachment.set_payload(pdf_content)
        self.message.attach(mimeattachment)

        self.message.send()
        data = self.get_api_call_json()
        attachments = data["Attachments"]
        self.assertEqual(len(attachments), 3)
        self.assertEqual(attachments[0]["Name"], "test.txt")
        self.assertEqual(attachments[0]["ContentType"], "text/plain")
        self.assertEqual(
            decode_att(attachments[0]["Content"]).decode("ascii"), text_content
        )
        self.assertNotIn("ContentID", attachments[0])

        # ContentType inferred from filename:
        self.assertEqual(attachments[1]["ContentType"], "image/png")
        self.assertEqual(attachments[1]["Name"], "test.png")
        self.assertEqual(decode_att(attachments[1]["Content"]), png_content)
        # make sure image not treated as inline:
        self.assertNotIn("ContentID", attachments[1])

        self.assertEqual(attachments[2]["ContentType"], "application/pdf")
        self.assertEqual(attachments[2]["Name"], "")  # none
        self.assertEqual(decode_att(attachments[2]["Content"]), pdf_content)
        self.assertNotIn("ContentID", attachments[2])

    def test_unicode_attachment_correctly_decoded(self):
        self.message.attach(
            "Une pièce jointe.html", "<p>\u2019</p>", mimetype="text/html"
        )
        self.message.send()
        data = self.get_api_call_json()
        self.assertEqual(
            data["Attachments"],
            [
                {
                    "Name": "Une pièce jointe.html",
                    "ContentType": "text/html",
                    "Content": b64encode("<p>\u2019</p>".encode("utf-8")).decode(
                        "ascii"
                    ),
                }
            ],
        )

    def test_embedded_images(self):
        image_filename = SAMPLE_IMAGE_FILENAME
        image_path = sample_image_path(image_filename)
        image_data = sample_image_content(image_filename)

        cid = attach_inline_image_file(self.message, image_path)  # Read from a png file
        html_content = (
            '<p>This has an <img src="cid:%s" alt="inline" /> image.</p>' % cid
        )
        self.message.attach_alternative(html_content, "text/html")

        self.message.send()
        data = self.get_api_call_json()
        self.assertEqual(data["HtmlBody"], html_content)

        attachments = data["Attachments"]
        self.assertEqual(len(attachments), 1)
        self.assertEqual(attachments[0]["Name"], image_filename)
        self.assertEqual(attachments[0]["ContentType"], "image/png")
        self.assertEqual(decode_att(attachments[0]["Content"]), image_data)
        self.assertEqual(attachments[0]["ContentID"], "cid:%s" % cid)

    def test_attached_images(self):
        image_filename = SAMPLE_IMAGE_FILENAME
        image_path = sample_image_path(image_filename)
        image_data = sample_image_content(image_filename)

        # option 1: attach as a file
        self.message.attach_file(image_path)

        # option 2: construct the MIMEImage and attach it directly
        image = MIMEImage(image_data)
        self.message.attach(image)

        image_data_b64 = b64encode(image_data).decode("ascii")

        self.message.send()
        data = self.get_api_call_json()
        self.assertEqual(
            data["Attachments"],
            [
                {
                    "Name": image_filename,  # the named one
                    "ContentType": "image/png",
                    "Content": image_data_b64,
                },
                {
                    "Name": "",  # the unnamed one
                    "ContentType": "image/png",
                    "Content": image_data_b64,
                },
            ],
        )

    def test_multiple_html_alternatives(self):
        # Multiple alternatives not allowed
        self.message.attach_alternative("<p>First html is OK</p>", "text/html")
        self.message.attach_alternative("<p>But not second html</p>", "text/html")
        with self.assertRaises(AnymailUnsupportedFeature):
            self.message.send()

    def test_html_alternative(self):
        # Only html alternatives allowed
        self.message.attach_alternative("{'not': 'allowed'}", "application/json")
        with self.assertRaises(AnymailUnsupportedFeature):
            self.message.send()

    def test_alternatives_fail_silently(self):
        # Make sure fail_silently is respected
        self.message.attach_alternative("{'not': 'allowed'}", "application/json")
        sent = self.message.send(fail_silently=True)
        self.assert_esp_not_called("API should not be called when send fails silently")
        self.assertEqual(sent, 0)

    def test_suppress_empty_address_lists(self):
        """Empty to, cc, bcc, and reply_to shouldn't generate empty fields"""
        self.message.send()
        data = self.get_api_call_json()
        self.assertNotIn("Cc", data)
        self.assertNotIn("Bcc", data)
        self.assertNotIn("ReplyTo", data)

        # Test empty `to`--but send requires at least one recipient somewhere (like cc)
        self.message.to = []
        self.message.cc = ["cc@example.com"]
        self.message.send()
        data = self.get_api_call_json()
        self.assertNotIn("To", data)

    def test_multiple_from_emails(self):
        """
        Postmark accepts multiple addresses in from_email (though only uses the first)
        """
        self.message.from_email = 'first@example.com, "From, also" <second@example.com>'
        self.message.send()
        data = self.get_api_call_json()
        self.assertEqual(
            data["From"], 'first@example.com, "From, also" <second@example.com>'
        )

        # Make sure the far-more-likely scenario of a single from_email
        # with an unquoted display-name issues a reasonable error:
        self.message.from_email = "Unquoted, display-name <from@example.com>"
        with self.assertRaises(AnymailInvalidAddress):
            self.message.send()

    def test_api_failure(self):
        self.set_mock_response(status_code=500)
        with self.assertRaisesMessage(AnymailAPIError, "Postmark API response 500"):
            mail.send_mail("Subject", "Body", "from@example.com", ["to@example.com"])

        # Make sure fail_silently is respected
        self.set_mock_response(status_code=500)
        sent = mail.send_mail(
            "Subject",
            "Body",
            "from@example.com",
            ["to@example.com"],
            fail_silently=True,
        )
        self.assertEqual(sent, 0)

    def test_api_error_includes_details(self):
        """AnymailAPIError should include ESP's error message"""
        # JSON error response:
        error_response = b"""{
            "ErrorCode": 451,
            "Message": "Helpful explanation from Postmark."
        }"""
        self.set_mock_response(status_code=200, raw=error_response)
        with self.assertRaisesMessage(
            AnymailAPIError, "Helpful explanation from Postmark"
        ):
            self.message.send()

        # Non-JSON error response:
        self.set_mock_response(status_code=500, raw=b"Ack! Bad proxy!")
        with self.assertRaisesMessage(AnymailAPIError, "Ack! Bad proxy!"):
            self.message.send()

        # No content in the error response:
        self.set_mock_response(status_code=502, raw=None)
        with self.assertRaises(AnymailAPIError):
            self.message.send()


@tag("postmark")
class PostmarkBackendAnymailFeatureTests(PostmarkBackendMockAPITestCase):
    """Test backend support for Anymail added features"""

    def test_envelope_sender(self):
        # Postmark doesn't allow overriding envelope sender on individual messages.
        # You can configure a custom return-path domain for each server in their
        # control panel.
        self.message.envelope_sender = "anything@bounces.example.com"
        with self.assertRaisesMessage(AnymailUnsupportedFeature, "envelope_sender"):
            self.message.send()

    def test_metadata(self):
        self.message.metadata = {"user_id": "12345", "items": 6}
        self.message.send()
        data = self.get_api_call_json()
        self.assertEqual(data["Metadata"], {"user_id": "12345", "items": 6})

    def test_send_at(self):
        self.message.send_at = 1651820889  # 2022-05-06 07:08:09 UTC
        with self.assertRaisesMessage(AnymailUnsupportedFeature, "send_at"):
            self.message.send()

    def test_tags(self):
        self.message.tags = ["receipt"]
        self.message.send()
        data = self.get_api_call_json()
        self.assertEqual(data["Tag"], "receipt")

        self.message.tags = ["receipt", "repeat-user"]
        with self.assertRaisesMessage(AnymailUnsupportedFeature, "multiple tags"):
            self.message.send()

    def test_track_opens(self):
        self.message.track_opens = True
        self.message.send()
        data = self.get_api_call_json()
        self.assertEqual(data["TrackOpens"], True)

    def test_track_clicks(self):
        self.message.track_clicks = True
        self.message.send()
        data = self.get_api_call_json()
        self.assertEqual(data["TrackLinks"], "HtmlAndText")

        # Also explicit "None" for False (to override server default)
        self.message.track_clicks = False
        self.message.send()
        data = self.get_api_call_json()
        self.assertEqual(data["TrackLinks"], "None")

    def test_template(self):
        message = AnymailMessage(
            # Omit subject and body (Postmark prohibits them with templates)
            from_email="from@example.com",
            to=["to@example.com"],
            template_id=1234567,
            merge_global_data={"name": "Alice", "group": "Developers"},
        )
        message.send()
        self.assert_esp_called("/email/withTemplate/")
        data = self.get_api_call_json()
        self.assertEqual(data["TemplateId"], 1234567)
        self.assertEqual(
            data["TemplateModel"], {"name": "Alice", "group": "Developers"}
        )
        # Make sure Django default subject and body didn't end up in the payload:
        self.assertNotIn("Subject", data)
        self.assertNotIn("HtmlBody", data)
        self.assertNotIn("TextBody", data)

    def test_template_alias(self):
        # Anymail template_id can be either Postmark TemplateId or TemplateAlias
        message = AnymailMessage(
            from_email="from@example.com",
            to=["to@example.com"],
            template_id="welcome-message",
        )
        message.send()
        self.assert_esp_called("/email/withTemplate/")
        data = self.get_api_call_json()
        self.assertEqual(data["TemplateAlias"], "welcome-message")
        # Postmark requires TemplateModel (can be empty) with TemplateId/TemplateAlias
        self.assertEqual(data["TemplateModel"], {})

    def test_template_multiple_recipients(self):
        # This is a non-batch (no merge_data) template send
        message = AnymailMessage(
            from_email="from@example.com",
            to=["to@example.com", "Also to <to2@example.com>"],
            template_id=1234567,
        )
        message.send()
        self.assert_esp_called("/email/withTemplate/")
        data = self.get_api_call_json()
        self.assertEqual(data["To"], "to@example.com, Also to <to2@example.com>")
        self.assertEqual(data["TemplateId"], 1234567)

    _mock_batch_response = json.dumps(
        [
            {
                "ErrorCode": 0,
                "Message": "OK",
                "To": "alice@example.com",
                "SubmittedAt": "2016-03-12T15:27:50.4468803-05:00",
                "MessageID": "b7bc2f4a-e38e-4336-af7d-e6c392c2f817",
            },
            {
                "ErrorCode": 0,
                "Message": "OK",
                "To": "bob@example.com",
                "SubmittedAt": "2016-03-12T15:27:50.4468803-05:00",
                "MessageID": "e2ecbbfc-fe12-463d-b933-9fe22915106d",
            },
        ]
    ).encode("utf-8")

    def test_merge_data(self):
        self.set_mock_response(raw=self._mock_batch_response)
        message = AnymailMessage(
            from_email="from@example.com",
            # Postmark only supports merge_data content in a template
            template_id=1234567,
            to=["alice@example.com", "Bob <bob@example.com>"],
            merge_data={
                "alice@example.com": {"name": "Alice", "group": "Developers"},
                "bob@example.com": {"name": "Bob"},  # and leave group undefined
                "nobody@example.com": {"name": "Not a recipient for this message"},
            },
            merge_global_data={"group": "Users", "site": "ExampleCo"},
        )
        message.send()

        self.assert_esp_called("/email/batchWithTemplates")
        data = self.get_api_call_json()
        messages = data["Messages"]
        self.assertEqual(len(messages), 2)
        self.assertEqual(
            messages[0],
            {
                "From": "from@example.com",
                "To": "alice@example.com",
                "TemplateId": 1234567,
                "TemplateModel": {
                    "name": "Alice",
                    "group": "Developers",
                    "site": "ExampleCo",
                },
            },
        )
        self.assertEqual(
            messages[1],
            {
                "From": "from@example.com",
                "To": "Bob <bob@example.com>",
                "TemplateId": 1234567,
                "TemplateModel": {"name": "Bob", "group": "Users", "site": "ExampleCo"},
            },
        )

        recipients = message.anymail_status.recipients
        self.assertEqual(recipients["alice@example.com"].status, "sent")
        self.assertEqual(
            recipients["alice@example.com"].message_id,
            "b7bc2f4a-e38e-4336-af7d-e6c392c2f817",
        )
        self.assertEqual(recipients["bob@example.com"].status, "sent")
        self.assertEqual(
            recipients["bob@example.com"].message_id,
            "e2ecbbfc-fe12-463d-b933-9fe22915106d",
        )

    def test_merge_data_single_recipient(self):
        self.set_mock_response(raw=self._mock_batch_response)
        message = AnymailMessage(
            from_email="from@example.com",
            # Postmark only supports merge_data content in a template:
            template_id=1234567,
            to=["alice@example.com"],
            merge_data={
                "alice@example.com": {"name": "Alice", "group": "Developers"},
                "nobody@example.com": {"name": "Not a recipient for this message"},
            },
            merge_global_data={"group": "Users", "site": "ExampleCo"},
        )
        message.send()

        # because merge_data is set, it's treated as a batch send
        self.assert_esp_called("/email/batchWithTemplates")
        data = self.get_api_call_json()

        self.assertEqual(
            data,
            {
                "Messages": [
                    {
                        "From": "from@example.com",
                        "To": "alice@example.com",
                        "TemplateId": 1234567,
                        "TemplateModel": {
                            "name": "Alice",
                            "group": "Developers",
                            "site": "ExampleCo",
                        },
                    }
                ]
            },
        )

        recipients = message.anymail_status.recipients
        self.assertEqual(recipients["alice@example.com"].status, "sent")
        self.assertEqual(
            recipients["alice@example.com"].message_id,
            "b7bc2f4a-e38e-4336-af7d-e6c392c2f817",
        )

    def test_merge_data_no_template(self):
        # merge_data={} can be used to force batch sending without a template
        self.set_mock_response(raw=self._mock_batch_response)
        message = AnymailMessage(
            from_email="from@example.com",
            to=["alice@example.com", "Bob <bob@example.com>"],
            merge_data={},
            subject="Test batch send",
            body="Test body",
        )
        message.send()

        self.assert_esp_called("/email/batch")
        data = self.get_api_call_json()
        self.assertEqual(len(data), 2)
        self.assertEqual(
            data[0],
            {
                "From": "from@example.com",
                "To": "alice@example.com",
                "Subject": "Test batch send",
                "TextBody": "Test body",
            },
        )
        self.assertEqual(
            data[1],
            {
                "From": "from@example.com",
                "To": "Bob <bob@example.com>",
                "Subject": "Test batch send",
                "TextBody": "Test body",
            },
        )

        recipients = message.anymail_status.recipients
        self.assertEqual(recipients["alice@example.com"].status, "sent")
        self.assertEqual(
            recipients["alice@example.com"].message_id,
            "b7bc2f4a-e38e-4336-af7d-e6c392c2f817",
        )
        self.assertEqual(recipients["bob@example.com"].status, "sent")
        self.assertEqual(
            recipients["bob@example.com"].message_id,
            "e2ecbbfc-fe12-463d-b933-9fe22915106d",
        )

    def test_merge_metadata(self):
        self.set_mock_response(raw=self._mock_batch_response)
        self.message.to = ["alice@example.com", "Bob <bob@example.com>"]
        self.message.merge_metadata = {
            "alice@example.com": {"order_id": 123, "tier": "premium"},
            "bob@example.com": {"order_id": 678},
        }
        self.message.metadata = {"notification_batch": "zx912"}
        self.message.send()

        self.assert_esp_called("/email/batch")
        data = self.get_api_call_json()
        self.assertEqual(len(data), 2)
        self.assertEqual(data[0]["To"], "alice@example.com")
        # metadata and merge_metadata[recipient] are combined:
        self.assertEqual(
            data[0]["Metadata"],
            {"order_id": 123, "tier": "premium", "notification_batch": "zx912"},
        )
        self.assertEqual(data[1]["To"], "Bob <bob@example.com>")
        self.assertEqual(
            data[1]["Metadata"], {"order_id": 678, "notification_batch": "zx912"}
        )

    def test_merge_metadata_with_template(self):
        self.set_mock_response(raw=self._mock_batch_response)
        self.message.to = ["alice@example.com", "Bob <bob@example.com>"]
        self.message.template_id = 1234567
        self.message.merge_metadata = {
            "alice@example.com": {"order_id": 123},
            "bob@example.com": {"order_id": 678, "tier": "premium"},
        }
        self.message.send()

        self.assert_esp_called("/email/batchWithTemplates")
        data = self.get_api_call_json()
        messages = data["Messages"]
        self.assertEqual(len(messages), 2)
        self.assertEqual(messages[0]["To"], "alice@example.com")
        # metadata and merge_metadata[recipient] are combined:
        self.assertEqual(messages[0]["Metadata"], {"order_id": 123})
        self.assertEqual(messages[1]["To"], "Bob <bob@example.com>")
        self.assertEqual(messages[1]["Metadata"], {"order_id": 678, "tier": "premium"})

    def test_merge_headers(self):
        self.message.to = ["alice@example.com", "Bob <bob@example.com>"]
        self.message.extra_headers = {
            "List-Unsubscribe-Post": "List-Unsubscribe=One-Click",
            "List-Unsubscribe": "<mailto:unsubscribe@example.com>",
        }
        self.message.merge_headers = {
            "alice@example.com": {
                "List-Unsubscribe": "<https://example.com/a/>",
            },
            "bob@example.com": {
                "List-Unsubscribe": "<https://example.com/b/>",
            },
        }
        self.message.send()

        self.assert_esp_called("/email/batch")
        data = self.get_api_call_json()
        self.assertEqual(len(data), 2)
        # Global and merge headers are combined:
        self.assertEqual(data[0]["To"], "alice@example.com")
        self.assertCountEqual(
            data[0]["Headers"],
            [
                {"Name": "List-Unsubscribe", "Value": "<https://example.com/a/>"},
                {
                    "Name": "List-Unsubscribe-Post",
                    "Value": "List-Unsubscribe=One-Click",
                },
            ],
        )
        self.assertEqual(data[1]["To"], "Bob <bob@example.com>")
        self.assertCountEqual(
            data[1]["Headers"],
            [
                {"Name": "List-Unsubscribe", "Value": "<https://example.com/b/>"},
                {
                    "Name": "List-Unsubscribe-Post",
                    "Value": "List-Unsubscribe=One-Click",
                },
            ],
        )

    def test_default_omits_options(self):
        """Make sure by default we don't send any ESP-specific options.

        Options not specified by the caller should be omitted entirely from
        the API call (*not* sent as False or empty). This ensures
        that your ESP account settings apply by default.
        """
        self.message.send()
        data = self.get_api_call_json()
        self.assertNotIn("Metadata", data)
        self.assertNotIn("Tag", data)
        self.assertNotIn("TemplateId", data)
        self.assertNotIn("TemplateModel", data)
        self.assertNotIn("TrackOpens", data)
        self.assertNotIn("TrackLinks", data)

    def test_esp_extra(self):
        self.message.esp_extra = {
            "FuturePostmarkOption": "some-value",
        }
        self.message.send()
        data = self.get_api_call_json()
        self.assertEqual(data["FuturePostmarkOption"], "some-value")

    def test_message_server_token(self):
        # Can override server-token on a per-message basis:
        self.message.esp_extra = {
            "server_token": "token_for_this_message_only",
        }
        self.message.send()
        headers = self.get_api_call_headers()
        self.assertEqual(
            headers["X-Postmark-Server-Token"], "token_for_this_message_only"
        )
        data = self.get_api_call_json()
        self.assertNotIn("server_token", data)  # not in the json

    # noinspection PyUnresolvedReferences
    def test_send_attaches_anymail_status(self):
        """The anymail_status should be attached to the message when it is sent"""
        response_content = b"""{
            "MessageID":"abcdef01-2345-6789-0123-456789abcdef",
            "ErrorCode":0,
            "To":"Recipient <to1@example.com>",
            "Message":"OK"
        }"""
        self.set_mock_response(raw=response_content)
        msg = mail.EmailMessage(
            "Subject",
            "Message",
            "from@example.com",
            ["Recipient <to1@example.com>"],
        )
        sent = msg.send()
        self.assertEqual(sent, 1)
        self.assertEqual(msg.anymail_status.status, {"sent"})
        self.assertEqual(
            msg.anymail_status.message_id, "abcdef01-2345-6789-0123-456789abcdef"
        )
        self.assertEqual(
            msg.anymail_status.recipients["to1@example.com"].status, "sent"
        )
        self.assertEqual(
            msg.anymail_status.recipients["to1@example.com"].message_id,
            "abcdef01-2345-6789-0123-456789abcdef",
        )
        self.assertEqual(msg.anymail_status.esp_response.content, response_content)

    # noinspection PyUnresolvedReferences
    def test_send_without_to_attaches_anymail_status(self):
        """The anymail_status should be attached even if there are no `to` recipients"""
        # Despite Postmark's docs, the "To" field is *not* required
        # if cc or bcc is provided.
        response_content = b"""{
            "SubmittedAt": "2019-01-28T13:54:35.5813997-05:00",
            "MessageID":"abcdef01-2345-6789-0123-456789abcdef",
            "ErrorCode":0,
            "Message":"OK"
        }"""
        self.set_mock_response(raw=response_content)
        msg = mail.EmailMessage(
            "Subject",
            "Message",
            "from@example.com",
            cc=["cc@example.com"],
        )
        sent = msg.send()
        self.assertEqual(sent, 1)
        self.assertEqual(msg.anymail_status.status, {"sent"})
        self.assertEqual(
            msg.anymail_status.message_id, "abcdef01-2345-6789-0123-456789abcdef"
        )
        self.assertEqual(msg.anymail_status.recipients["cc@example.com"].status, "sent")
        self.assertEqual(
            msg.anymail_status.recipients["cc@example.com"].message_id,
            "abcdef01-2345-6789-0123-456789abcdef",
        )
        self.assertEqual(msg.anymail_status.esp_response.content, response_content)

    # noinspection PyUnresolvedReferences
    def test_send_failed_anymail_status(self):
        """If the send fails, anymail_status should contain initial values"""
        self.set_mock_response(status_code=500)
        sent = self.message.send(fail_silently=True)
        self.assertEqual(sent, 0)
        self.assertIsNone(self.message.anymail_status.status)
        self.assertIsNone(self.message.anymail_status.message_id)
        self.assertEqual(self.message.anymail_status.recipients, {})
        self.assertIsNone(self.message.anymail_status.esp_response)

    # noinspection PyUnresolvedReferences
    def test_send_unparsable_response(self):
        """
        If the send succeeds, but a non-JSON API response, should raise an API exception
        """
        mock_response = self.set_mock_response(
            status_code=200, raw=b"yikes, this isn't a real response"
        )
        with self.assertRaises(AnymailAPIError):
            self.message.send()
        self.assertIsNone(self.message.anymail_status.status)
        self.assertIsNone(self.message.anymail_status.message_id)
        self.assertEqual(self.message.anymail_status.recipients, {})
        self.assertEqual(self.message.anymail_status.esp_response, mock_response)

    def test_json_serialization_errors(self):
        """Try to provide more information about non-json-serializable data"""
        self.message.tags = [Decimal("19.99")]  # yeah, don't do this
        with self.assertRaises(AnymailSerializationError) as cm:
            self.message.send()
            print(self.get_api_call_json())
        err = cm.exception
        self.assertIsInstance(err, TypeError)  # compatibility with json.dumps
        # our added context:
        self.assertIn("Don't know how to send this data to Postmark", str(err))
        # original message:
        self.assertRegex(str(err), r"Decimal.*is not JSON serializable")


@tag("postmark")
class PostmarkBackendRecipientsRefusedTests(PostmarkBackendMockAPITestCase):
    """
    Should raise AnymailRecipientsRefused when *all* recipients are rejected or invalid
    """

    def test_recipients_inactive(self):
        self.set_mock_response(
            status_code=422,
            raw=b'{"ErrorCode":406,'
            b'"Message":"You tried to send to a recipient'
            b" that has been marked as inactive.\\n"
            b"Found inactive addresses: hardbounce@example.com, spam@example.com.\\n"
            b"Inactive recipients are ones that have generated"
            b' a hard bounce or a spam complaint."}',
        )
        msg = mail.EmailMessage(
            "Subject",
            "Body",
            "from@example.com",
            ["HardBounce@example.com", "Hates Spam <spam@example.com>"],
        )
        with self.assertRaises(AnymailRecipientsRefused):
            msg.send()
        status = msg.anymail_status
        self.assertEqual(status.recipients["HardBounce@example.com"].status, "rejected")
        self.assertEqual(status.recipients["spam@example.com"].status, "rejected")

    def test_recipients_invalid(self):
        self.set_mock_response(
            status_code=422,
            raw=b"""{
                "ErrorCode":300,
                "Message":"Invalid 'To' address: 'invalid@localhost'."
            }""",
        )
        msg = mail.EmailMessage(
            "Subject", "Body", "from@example.com", ["Invalid@LocalHost"]
        )
        with self.assertRaises(AnymailRecipientsRefused):
            msg.send()
        status = msg.anymail_status
        self.assertEqual(status.recipients["Invalid@LocalHost"].status, "invalid")

    def test_recipients_parse_error(self):
        self.set_mock_response(
            status_code=422,
            raw=b"""{
              "ErrorCode": 300,
              "Message": "Error parsing 'Cc': Illegal email domain '+' in address 'user@+'."
            }""",  # NOQA: E501
        )
        msg = mail.EmailMessage("Subject", "Body", "from@example.com", cc=["user@+"])
        with self.assertRaises(AnymailRecipientsRefused):
            msg.send()
        status = msg.anymail_status
        self.assertEqual(status.recipients["user@+"].status, "invalid")

    def test_from_email_invalid(self):
        # Invalid 'From' address generates same Postmark ErrorCode 300 as invalid 'To',
        # but should raise a different Anymail error
        self.set_mock_response(
            status_code=422,
            raw=b"""{
                "ErrorCode":300,
                "Message":"Invalid 'From' address: 'invalid@localhost'."
            }""",
        )
        msg = mail.EmailMessage(
            "Subject", "Body", "invalid@localhost", ["to@example.com"]
        )
        with self.assertRaises(AnymailAPIError):
            msg.send()

    def test_reply_to_invalid(self):
        # Make sure 'Reply-To' error doesn't get caught in invalid 'To' logic:
        self.set_mock_response(
            status_code=422,
            raw=b"""{
                "ErrorCode": 300,
                "Message": "Error parsing 'Reply-To': Illegal email domain '+' in address 'invalid@+'."}
            """,  # NOQA: E501
        )
        msg = mail.EmailMessage(
            "Subject",
            "Body",
            "from@example.com",
            ["to@example.com"],
            reply_to=["invalid@+"],
        )
        with self.assertRaisesMessage(AnymailAPIError, "Error parsing 'Reply-To'"):
            msg.send()

    def test_errorcode_300(self):
        # Various other problems generate same Postmark ErrorCode 300 as invalid 'To',
        # but should raise ordinary API errors
        self.set_mock_response(
            status_code=422,
            raw=b"""{
                "ErrorCode": 300,
                "Message": "Invalid metadata content. Field names are limited to 20 characters..."
            }""",  # NOQA: E501
        )
        msg = mail.EmailMessage(
            "Subject", "Body", "from@example.com", ["to@example.com"]
        )
        msg.metadata = {"this-key-name-is-too-long": "data"}
        with self.assertRaisesMessage(AnymailAPIError, "Invalid metadata content"):
            msg.send()

    def test_fail_silently(self):
        self.set_mock_response(
            status_code=422,
            raw=b'{"ErrorCode":406,'
            b'"Message":"You tried to send to a recipient'
            b" that has been marked as inactive.\\n"
            b"Found inactive addresses: hardbounce@example.com, spam@example.com.\\n"
            b"Inactive recipients are ones that have generated"
            b' a hard bounce or a spam complaint."}',
        )
        msg = mail.EmailMessage(
            "Subject",
            "Body",
            "from@example.com",
            ["HardBounce@example.com", "Hates Spam <spam@example.com>"],
        )
        msg.send(fail_silently=True)
        status = msg.anymail_status
        self.assertEqual(status.recipients["HardBounce@example.com"].status, "rejected")
        self.assertEqual(status.recipients["spam@example.com"].status, "rejected")

    @override_settings(ANYMAIL_IGNORE_RECIPIENT_STATUS=True)
    def test_ignore_recipient_status(self):
        self.set_mock_response(
            status_code=422,
            raw=b'{"ErrorCode":406,'
            b'"Message":"You tried to send to a recipient'
            b" that has been marked as inactive.\\n"
            b"Found inactive addresses: hardbounce@example.com, spam@example.com.\\n"
            b"Inactive recipients are ones that have generated"
            b' a hard bounce or a spam complaint. "}',
        )
        msg = mail.EmailMessage(
            "Subject",
            "Body",
            "from@example.com",
            ["HardBounce@example.com", "Hates Spam <spam@example.com>"],
        )
        msg.send()
        status = msg.anymail_status
        self.assertEqual(status.recipients["HardBounce@example.com"].status, "rejected")
        self.assertEqual(status.recipients["spam@example.com"].status, "rejected")

    def test_mixed_response(self):
        """If *any* recipients are valid or queued, no exception is raised"""
        self.set_mock_response(
            status_code=200,
            raw=b'{"To":"hardbounce@example.com, valid@example.com,'
            b' Hates Spam <spam@example.com>",'
            b'"SubmittedAt":"2016-03-12T22:59:06.2505871-05:00",'
            b'"MessageID":"089dce03-feee-408e-9f0c-ee69bf1c5f35",'
            b'"ErrorCode":0,'
            b'"Message":"Message OK, but will not deliver to these inactive addresses:'
            b" hardbounce@example.com, spam@example.com."
            b" Inactive recipients are ones that have generated"
            b' a hard bounce or a spam complaint."}',
        )
        msg = mail.EmailMessage(
            "Subject",
            "Body",
            "from@example.com",
            [
                "HardBounce@example.com",
                "valid@example.com",
                "Hates Spam <spam@example.com>",
            ],
        )
        sent = msg.send()
        # one message sent, successfully, to 1 of 3 recipients:
        self.assertEqual(sent, 1)
        status = msg.anymail_status
        self.assertEqual(status.recipients["HardBounce@example.com"].status, "rejected")
        self.assertEqual(status.recipients["valid@example.com"].status, "sent")
        self.assertEqual(status.recipients["spam@example.com"].status, "rejected")

    def test_delivery_delayed_response(self):
        # Postmark's "delivery may be delayed" response doesn't have an ErrorCode
        # field set to 0 (despite their API docs).
        response_data = {
            "Message": "Message accepted, but delivery may be delayed.",
            "MessageID": "38360f97-ff7f-44b2-bcd1-5ea94ff2af00",
            "SubmittedAt": "2024-08-05T02:03:37.0951168Z",
            "To": "to@example.com",
        }
        self.set_mock_response(json_data=response_data)
        sent = self.message.send()
        self.assertEqual(sent, 1)
        status = self.message.anymail_status
        self.assertEqual(status.recipients["to@example.com"].status, "queued")
        self.assertEqual(status.message_id, "38360f97-ff7f-44b2-bcd1-5ea94ff2af00")


@tag("postmark")
class PostmarkBackendSessionSharingTestCase(
    SessionSharingTestCases, PostmarkBackendMockAPITestCase
):
    """Requests session sharing tests"""

    pass  # tests are defined in SessionSharingTestCases


@tag("postmark")
@override_settings(EMAIL_BACKEND="anymail.backends.postmark.EmailBackend")
class PostmarkBackendImproperlyConfiguredTests(AnymailTestMixin, SimpleTestCase):
    """Test ESP backend without required settings in place"""

    def test_missing_api_key(self):
        with self.assertRaises(ImproperlyConfigured) as cm:
            mail.send_mail("Subject", "Message", "from@example.com", ["to@example.com"])
        errmsg = str(cm.exception)
        self.assertRegex(errmsg, r"\bPOSTMARK_SERVER_TOKEN\b")
        self.assertRegex(errmsg, r"\bANYMAIL_POSTMARK_SERVER_TOKEN\b")



import os
import unittest
from datetime import datetime, timedelta
from email.utils import formataddr

from django.test import SimpleTestCase, override_settings, tag

from anymail.exceptions import AnymailAPIError
from anymail.message import AnymailMessage

from .utils import AnymailTestMixin, sample_image_path

ANYMAIL_TEST_SPARKPOST_API_KEY = os.getenv("ANYMAIL_TEST_SPARKPOST_API_KEY")
ANYMAIL_TEST_SPARKPOST_DOMAIN = os.getenv("ANYMAIL_TEST_SPARKPOST_DOMAIN")


@tag("sparkpost", "live")
@unittest.skipUnless(
    ANYMAIL_TEST_SPARKPOST_API_KEY and ANYMAIL_TEST_SPARKPOST_DOMAIN,
    "Set ANYMAIL_TEST_SPARKPOST_API_KEY and ANYMAIL_TEST_SPARKPOST_DOMAIN "
    "environment variables to run SparkPost integration tests",
)
@override_settings(
    ANYMAIL_SPARKPOST_API_KEY=ANYMAIL_TEST_SPARKPOST_API_KEY,
    EMAIL_BACKEND="anymail.backends.sparkpost.EmailBackend",
)
class SparkPostBackendIntegrationTests(AnymailTestMixin, SimpleTestCase):
    """SparkPost API integration tests

    These tests run against the **live** SparkPost API, using the
    environment variable `ANYMAIL_TEST_SPARKPOST_API_KEY` as the API key
    If that variable is not set, these tests won't run.

    SparkPost doesn't offer a test mode -- it tries to send everything
    you ask. To avoid stacking up a pile of undeliverable @example.com
    emails, the tests use SparkPost's "sink domain" @*.sink.sparkpostmail.com.
    https://www.sparkpost.com/docs/faq/using-sink-server/
    """

    def setUp(self):
        super().setUp()
        self.from_email = "test@%s" % ANYMAIL_TEST_SPARKPOST_DOMAIN
        self.message = AnymailMessage(
            "Anymail SparkPost integration test",
            "Text content",
            self.from_email,
            ["to@test.sink.sparkpostmail.com"],
        )
        self.message.attach_alternative("<p>HTML content</p>", "text/html")

    def test_simple_send(self):
        # Example of getting the SparkPost send status
        # and transmission id from the message
        sent_count = self.message.send()
        self.assertEqual(sent_count, 1)

        anymail_status = self.message.anymail_status
        sent_status = anymail_status.recipients["to@test.sink.sparkpostmail.com"].status
        message_id = anymail_status.recipients[
            "to@test.sink.sparkpostmail.com"
        ].message_id

        self.assertEqual(sent_status, "queued")  # SparkPost always queues
        # this is actually the transmission_id; should be non-blank:
        self.assertRegex(message_id, r".+")
        # set of all recipient statuses:
        self.assertEqual(anymail_status.status, {sent_status})
        self.assertEqual(anymail_status.message_id, message_id)

    def test_all_options(self):
        send_at = datetime.now() + timedelta(minutes=2)
        message = AnymailMessage(
            subject="Anymail all-options integration test",
            body="This is the text body",
            from_email=formataddr(("Test From, with comma", self.from_email)),
            to=[
                "to1@test.sink.sparkpostmail.com",
                "Recipient 2 <to2@test.sink.sparkpostmail.com>",
            ],
            # Limit the live b/cc's to avoid running through our small monthly
            # allowance:
            cc=["Copy To <cc@test.sink.sparkpostmail.com>"],
            reply_to=["reply1@example.com", "Reply 2 <reply2@example.com>"],
            headers={"X-Anymail-Test": "value"},
            metadata={"meta1": "simple string", "meta2": 2},
            send_at=send_at,
            tags=["tag 1"],  # SparkPost only supports single tags
            track_clicks=True,
            track_opens=True,
        )
        message.attach("attachment1.txt", "Here is some\ntext for you", "text/plain")
        message.attach("attachment2.csv", "ID,Name\n1,Amy Lina", "text/csv")
        cid = message.attach_inline_image_file(sample_image_path())
        message.attach_alternative(
            "<p><b>HTML:</b> with <a href='http://example.com'>link</a>"
            "and image: <img src='cid:%s'></div>" % cid,
            "text/html",
        )

        message.send()
        # SparkPost always queues:
        self.assertEqual(message.anymail_status.status, {"queued"})

    def test_merge_data(self):
        message = AnymailMessage(
            subject="Anymail merge_data test: {{ value }}",
            body="This body includes merge data: {{ value }}\n"
            "And global merge data: {{ global }}",
            from_email=formataddr(("Test From", self.from_email)),
            to=[
                "to1@test.sink.sparkpostmail.com",
                "Recipient 2 <to2@test.sink.sparkpostmail.com>",
            ],
            merge_data={
                "to1@test.sink.sparkpostmail.com": {"value": "one"},
                "to2@test.sink.sparkpostmail.com": {"value": "two"},
            },
            merge_global_data={"global": "global_value"},
            merge_metadata={
                "to1@test.sink.sparkpostmail.com": {"meta1": "one"},
                "to2@test.sink.sparkpostmail.com": {"meta1": "two"},
            },
            headers={
                "X-Custom": "custom header default",
            },
            merge_headers={
                # (Note that SparkPost doesn't support custom List-Unsubscribe headers)
                "to1@test.sink.sparkpostmail.com": {
                    "X-Custom": "custom header one",
                },
            },
        )
        message.send()
        recipient_status = message.anymail_status.recipients
        self.assertEqual(
            recipient_status["to1@test.sink.sparkpostmail.com"].status, "queued"
        )
        self.assertEqual(
            recipient_status["to2@test.sink.sparkpostmail.com"].status, "queued"
        )

    def test_stored_template(self):
        message = AnymailMessage(
            # a real template in our SparkPost test account:
            template_id="test-template",
            to=["to1@test.sink.sparkpostmail.com"],
            merge_data={
                "to1@test.sink.sparkpostmail.com": {
                    "name": "Test Recipient",
                }
            },
            merge_global_data={
                "order": "12345",
            },
        )
        message.from_email = None  # from_email must come from stored template
        message.send()
        recipient_status = message.anymail_status.recipients
        self.assertEqual(
            recipient_status["to1@test.sink.sparkpostmail.com"].status, "queued"
        )

    @override_settings(ANYMAIL_SPARKPOST_API_KEY="Hey, that's not an API key!")
    def test_invalid_api_key(self):
        with self.assertRaises(AnymailAPIError) as cm:
            self.message.send()
        err = cm.exception
        self.assertEqual(err.status_code, 401)
        # Make sure the exception message includes SparkPost's response:
        self.assertIn("Unauthorized", str(err))



import os
import unittest
from email.utils import formataddr

from django.test import SimpleTestCase, override_settings, tag

from anymail.exceptions import AnymailAPIError
from anymail.message import AnymailMessage

from .utils import AnymailTestMixin

ANYMAIL_TEST_RESEND_API_KEY = os.getenv("ANYMAIL_TEST_RESEND_API_KEY")
ANYMAIL_TEST_RESEND_DOMAIN = os.getenv("ANYMAIL_TEST_RESEND_DOMAIN")


@tag("resend", "live")
@unittest.skipUnless(
    ANYMAIL_TEST_RESEND_API_KEY and ANYMAIL_TEST_RESEND_DOMAIN,
    "Set ANYMAIL_TEST_RESEND_API_KEY and ANYMAIL_TEST_RESEND_DOMAIN "
    "environment variables to run Resend integration tests",
)
@override_settings(
    ANYMAIL_RESEND_API_KEY=ANYMAIL_TEST_RESEND_API_KEY,
    EMAIL_BACKEND="anymail.backends.resend.EmailBackend",
)
class ResendBackendIntegrationTests(AnymailTestMixin, SimpleTestCase):
    """Resend.com API integration tests

    Resend doesn't have sandbox so these tests run
    against the **live** Resend API, using the
    environment variable `ANYMAIL_TEST_RESEND_API_KEY` as the API key,
    and `ANYMAIL_TEST_RESEND_DOMAIN` to construct sender addresses.
    If those variables are not set, these tests won't run.

    """

    def setUp(self):
        super().setUp()
        self.from_email = "from@%s" % ANYMAIL_TEST_RESEND_DOMAIN
        self.message = AnymailMessage(
            "Anymail Resend integration test",
            "Text content",
            self.from_email,
            ["test+to1@anymail.dev"],
        )
        self.message.attach_alternative("<p>HTML content</p>", "text/html")

    def test_simple_send(self):
        # Example of getting the Resend message id from the message
        sent_count = self.message.send()
        self.assertEqual(sent_count, 1)

        anymail_status = self.message.anymail_status
        sent_status = anymail_status.recipients["test+to1@anymail.dev"].status
        message_id = anymail_status.recipients["test+to1@anymail.dev"].message_id

        self.assertEqual(sent_status, "queued")  # Resend always queues
        self.assertGreater(len(message_id), 0)  # non-empty string
        # set of all recipient statuses:
        self.assertEqual(anymail_status.status, {sent_status})
        self.assertEqual(anymail_status.message_id, message_id)

    def test_all_options(self):
        message = AnymailMessage(
            subject="Anymail Resend all-options integration test",
            body="This is the text body",
            # Verify workarounds for address formatting issues:
            from_email=formataddr(("Test «Från», med komma", self.from_email)),
            to=["test+to1@anymail.dev", '"Recipient 2, OK?" <test+to2@anymail.dev>'],
            cc=["test+cc1@anymail.dev", "Copy 2 <test+cc2@anymail.dev>"],
            bcc=["test+bcc1@anymail.dev", "Blind Copy 2 <test+bcc2@anymail.dev>"],
            reply_to=['"Reply, with comma" <reply@example.com>', "reply2@example.com"],
            headers={"X-Anymail-Test": "value", "X-Anymail-Count": 3},
            metadata={"meta1": "simple string", "meta2": 2},
            tags=["tag 1", "tag 2"],
            # Resend supports send_at or attachments, but not both at once.
            # send_at=datetime.now() + timedelta(minutes=2),
        )
        message.attach_alternative("<p>HTML content</p>", "text/html")

        message.attach("attachment1.txt", "Here is some\ntext for you", "text/plain")
        message.attach("attachment2.csv", "ID,Name\n1,Amy Lina", "text/csv")

        message.send()
        # Resend always queues:
        self.assertEqual(message.anymail_status.status, {"queued"})
        self.assertGreater(
            len(message.anymail_status.message_id), 0
        )  # non-empty string

    def test_batch_send(self):
        # merge_metadata, merge_headers, or merge_data will use batch send API
        message = AnymailMessage(
            subject="Anymail Resend batch sendintegration test",
            body="This is the text body",
            from_email=self.from_email,
            to=["test+to1@anymail.dev", '"Recipient 2" <test+to2@anymail.dev>'],
            metadata={"meta1": "simple string", "meta2": 2},
            merge_metadata={
                "test+to1@anymail.dev": {"meta3": "recipient 1"},
                "test+to2@anymail.dev": {"meta3": "recipient 2"},
            },
            tags=["tag 1", "tag 2"],
            headers={
                "List-Unsubscribe-Post": "List-Unsubscribe=One-Click",
                "List-Unsubscribe": "<mailto:unsubscribe@example.com>",
            },
            merge_headers={
                "test+to1@anymail.dev": {
                    "List-Unsubscribe": "<https://example.com/a/>",
                },
                "test+to2@anymail.dev": {
                    "List-Unsubscribe": "<https://example.com/b/>",
                },
            },
        )
        message.attach_alternative("<p>HTML content</p>", "text/html")
        # Resend does not support attachments in batch send (2024-11-21)
        # message.attach("attachment1.txt", "Here is some\ntext for you", "text/plain")

        message.send()
        # Resend always queues:
        self.assertEqual(message.anymail_status.status, {"queued"})
        recipient_status = message.anymail_status.recipients
        self.assertEqual(recipient_status["test+to1@anymail.dev"].status, "queued")
        self.assertEqual(recipient_status["test+to2@anymail.dev"].status, "queued")
        self.assertRegex(recipient_status["test+to1@anymail.dev"].message_id, r".+")
        self.assertRegex(recipient_status["test+to2@anymail.dev"].message_id, r".+")
        # Each recipient gets their own message_id:
        self.assertNotEqual(
            recipient_status["test+to1@anymail.dev"].message_id,
            recipient_status["test+to2@anymail.dev"].message_id,
        )

    @unittest.skip("Resend has stopped responding to bad/missing API keys (12/2023)")
    @override_settings(ANYMAIL_RESEND_API_KEY="Hey, that's not an API key!")
    def test_invalid_api_key(self):
        with self.assertRaisesMessage(AnymailAPIError, "API key is invalid"):
            self.message.send()



import os
import unittest
from datetime import datetime, timedelta
from email.utils import formataddr

from django.test import SimpleTestCase, override_settings, tag

from anymail.exceptions import AnymailAPIError
from anymail.message import AnymailMessage

from .utils import AnymailTestMixin, sample_image_path

ANYMAIL_TEST_MAILERSEND_API_TOKEN = os.getenv("ANYMAIL_TEST_MAILERSEND_API_TOKEN")
ANYMAIL_TEST_MAILERSEND_DOMAIN = os.getenv("ANYMAIL_TEST_MAILERSEND_DOMAIN")


@tag("mailersend", "live")
@unittest.skipUnless(
    ANYMAIL_TEST_MAILERSEND_API_TOKEN and ANYMAIL_TEST_MAILERSEND_DOMAIN,
    "Set ANYMAIL_TEST_MAILERSEND_API_TOKEN and ANYMAIL_TEST_MAILERSEND_DOMAIN"
    " environment variables to run MailerSend integration tests",
)
@override_settings(
    ANYMAIL={
        "MAILERSEND_API_TOKEN": ANYMAIL_TEST_MAILERSEND_API_TOKEN,
    },
    EMAIL_BACKEND="anymail.backends.mailersend.EmailBackend",
)
class MailerSendBackendIntegrationTests(AnymailTestMixin, SimpleTestCase):
    """MailerSend API integration tests

    These tests run against the **live** MailerSend API, using the
    environment variable `ANYMAIL_TEST_MAILERSEND_API_TOKEN` as the API token
    and `ANYMAIL_TEST_MAILERSEND_DOMAIN` as the sender domain.
    If those variables are not set, these tests won't run.

    """

    def setUp(self):
        super().setUp()
        self.from_email = f"from@{ANYMAIL_TEST_MAILERSEND_DOMAIN}"
        self.message = AnymailMessage(
            "Anymail MailerSend integration test",
            "Text content",
            self.from_email,
            ["test+to1@anymail.dev"],
        )
        self.message.attach_alternative("<p>HTML content</p>", "text/html")

    def test_simple_send(self):
        # Example of getting the MailerSend send status and message id from the message
        sent_count = self.message.send()
        self.assertEqual(sent_count, 1)

        anymail_status = self.message.anymail_status
        sent_status = anymail_status.recipients["test+to1@anymail.dev"].status
        message_id = anymail_status.recipients["test+to1@anymail.dev"].message_id

        self.assertEqual(sent_status, "queued")  # MailerSend queues
        # don't know what it'll be, but it should exist (and not be a bulk id):
        self.assertGreater(len(message_id), 0)
        self.assertFalse(message_id.startswith("bulk:"))

        # set of all recipient statuses:
        self.assertEqual(anymail_status.status, {sent_status})
        self.assertEqual(anymail_status.message_id, message_id)

    def test_all_options(self):
        send_at = datetime.now() + timedelta(minutes=2)
        from_email = formataddr(("Test From, with comma", self.from_email))
        message = AnymailMessage(
            subject="Anymail MailerSend all-options integration test",
            body="This is the text body",
            from_email=from_email,
            to=[
                "test+to1@anymail.dev",
                "Recipient 2 <test+to2@anymail.dev>",
                "test+bounce@anymail.dev",  # will be rejected
            ],
            cc=["test+cc1@anymail.dev", "Copy 2 <test+cc2@anymail.dev>"],
            bcc=["test+bcc1@anymail.dev", "Blind Copy 2 <test+bcc2@anymail.dev>"],
            # MailerSend only supports single reply_to:
            reply_to=["Reply <reply@example.com>"],
            headers={
                # Special handling for these headers (available to all accounts):
                "Precedence": "bulk",
                "In-Reply-To": "earlier-id@anymail.dev",
                # Only "enterprise accounts" can send other custom headers:
                "X-Custom": "anymail test",
            },
            send_at=send_at,
            tags=["tag 1", "tag 2"],
            track_clicks=False,
            track_opens=True,
        )
        message.attach("attachment1.txt", "Here is some\ntext for you", "text/plain")
        message.attach("vedhæftet fil.csv", "ID,Name\n1,3", "text/csv")
        cid = message.attach_inline_image_file(
            sample_image_path(), domain=ANYMAIL_TEST_MAILERSEND_DOMAIN
        )
        message.attach_alternative(
            f"<div>This is the <i>html</i> body <img src='cid:{cid}'></div>",
            "text/html",
        )

        message.send()

        # First two recipients should be queued, other should be bounced
        self.assertEqual(message.anymail_status.status, {"queued", "rejected"})
        recipient_status = message.anymail_status.recipients
        self.assertEqual(recipient_status["test+to1@anymail.dev"].status, "queued")
        self.assertEqual(recipient_status["test+to2@anymail.dev"].status, "queued")
        self.assertEqual(recipient_status["test+bounce@anymail.dev"].status, "rejected")

    def test_stored_template(self):
        message = AnymailMessage(
            # id of a real template named in Anymail's MailerSend test account:
            template_id="vywj2lpokkm47oqz",
            to=[
                "test+to1@anymail.dev",
                "test+to2@anymail.dev",
                "test+bounce@anymail.dev",  # will be rejected
            ],
            merge_data={
                "test+to1@anymail.dev": {"name": "First Recipient", "order": 12345},
                "test+to2@anymail.dev": {"name": "Second Recipient", "order": 67890},
                "test+bounce@anymail.dev": {"name": "Bounces", "order": 3},
            },
            merge_global_data={"date": "yesterday"},
            esp_extra={
                # CAREFUL: with "expose-to-list", all recipients will see
                #   every other recipient's email address. (See docs!)
                "batch_send_mode": "expose-to-list"
            },
        )
        message.from_email = None  # use template From
        message.send()
        recipient_status = message.anymail_status.recipients
        self.assertEqual(recipient_status["test+to1@anymail.dev"].status, "queued")
        self.assertEqual(recipient_status["test+to2@anymail.dev"].status, "queued")
        self.assertEqual(recipient_status["test+bounce@anymail.dev"].status, "rejected")
        self.assertFalse(
            recipient_status["test+to1@anymail.dev"].message_id.startswith("bulk:")
        )
        self.assertIsNone(recipient_status["test+bounce@anymail.dev"].message_id)

    def test_batch_send_mode_bulk(self):
        # Same test as above, but with batch_send_mode "use-bulk-email".
        # (Uses different API; status is handled very differently.)
        message = AnymailMessage(
            # id of a real template named in Anymail's MailerSend test account:
            template_id="vywj2lpokkm47oqz",
            to=[
                "test+to1@anymail.dev",
                "test+to2@anymail.dev",
                "test+bounce@anymail.dev",  # will be rejected
            ],
            merge_data={
                "test+to1@anymail.dev": {"name": "First Recipient", "order": 12345},
                "test+to2@anymail.dev": {"name": "Second Recipient", "order": 67890},
                "test+bounce@anymail.dev": {"name": "Bounces", "order": 3},
            },
            merge_global_data={"date": "yesterday"},
            esp_extra={"batch_send_mode": "use-bulk-email"},
        )
        message.from_email = None  # use template From
        message.send()
        recipient_status = message.anymail_status.recipients
        # With use-bulk-email, must poll bulk-email status API to determine status:
        self.assertEqual(recipient_status["test+to1@anymail.dev"].status, "unknown")
        self.assertEqual(recipient_status["test+to2@anymail.dev"].status, "unknown")
        self.assertEqual(recipient_status["test+bounce@anymail.dev"].status, "unknown")
        # With use-bulk-email, message_id will be MailerSend's bulk_email_id
        # rather than an actual message_id. Anymail adds "bulk:" to differentiate:
        self.assertTrue(
            recipient_status["test+to1@anymail.dev"].message_id.startswith("bulk:")
        )
        self.assertTrue(
            recipient_status["test+bounce@anymail.dev"].message_id.startswith("bulk:")
        )

    @override_settings(
        ANYMAIL={
            "MAILERSEND_API_TOKEN": "Hey, that's not an API token",
        }
    )
    def test_invalid_api_key(self):
        with self.assertRaisesMessage(AnymailAPIError, "Unauthenticated"):
            self.message.send()
